import uuid
from datetime import datetime
from typing import List, Optional
from uuid import UUID

from encord.orm.base_dto import BaseDTO, Field


class GetPresetParams(BaseDTO):
    top_level_folder_uuid: Optional[UUID] = Field(default=None, alias="topLevelFolderUuid")
    preset_uuids: Optional[List[UUID]] = Field(default=[], alias="uuids")
    page_token: Optional[str] = Field(default=None, alias="pageToken")
    page_size: Optional[int] = Field(default=None, alias="pageSize")


class FilterPreset(BaseDTO):
    uuid: uuid.UUID
    name: str
    description: Optional[str]
    created_at: Optional[datetime] = Field(default=None, alias="createdAt")
    last_updated_at: Optional[datetime] = Field(default=None, alias="lastUpdatedAt")


class FilterPresetDefinition(BaseDTO):
    local_filters: Optional[dict] = Field(default=None, alias="localFilters")
    global_filters: Optional[dict] = Field(default=None, alias="globalFilters")


class GetPresetsResponse(BaseDTO):
    results: List[FilterPreset]



"""
---
title: "Bundle (Batch)"
slug: "sdk-ref-bundle"
hidden: false
metadata:
  title: "Bundle"
  description: "Encord SDK Bundle (Batch)."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass, is_dataclass
from functools import reduce
from typing import Callable, ClassVar, Dict, Generator, Generic, List, Optional, Protocol, Type, TypeVar

from encord.http.limits import LABEL_ROW_BUNDLE_DEFAULT_LIMIT

log = logging.getLogger(__name__)

BundlablePayloadT = TypeVar("BundlablePayloadT", bound="BundlablePayload")


class BundlablePayload(Protocol[BundlablePayloadT]):
    """
    All payloads that work with bundles need to provide "add" method
    that would allow bundler to combine multiple payloads into one or few aggregates.
    """

    # This line ensures we're only allowing dataclasses for now
    __dataclass_fields__: ClassVar[Dict]

    def add(self, other: BundlablePayloadT) -> BundlablePayloadT: ...


T = TypeVar("T")
R = TypeVar("R")


@dataclass
class BundleResultHandler(Generic[T]):
    predicate: str
    handler: Callable[[T], None]


@dataclass
class BundleResultMapper(Generic[T]):
    result_mapping_predicate: Callable[[T], str]
    result_handler: BundleResultHandler[T]


class BundledOperation(Generic[BundlablePayloadT, R]):
    def __init__(
        self,
        operation: Callable[..., List[R]],
        result_mapper: Optional[Callable[[R], str]],
        limit: int,
    ) -> None:
        self.operation = operation
        self.result_mapper = result_mapper
        self.limit = limit
        self.payloads: List[BundlablePayloadT] = []
        self.result_handlers: Dict[str, Callable[[BundlablePayloadT], None]] = {}

    def append(self, payload: BundlablePayloadT, result_handler: Optional[BundleResultHandler]):
        self.payloads.append(payload)
        if result_handler is not None:
            self.result_handlers[result_handler.predicate] = result_handler.handler

    def get_bundled_payload(self) -> Generator[BundlablePayloadT, None, None]:
        for i in range(0, len(self.payloads), self.limit):
            yield reduce(lambda x, y: x.add(y), self.payloads[i : i + self.limit])


class Bundle:
    """
    This class allows to perform operations in bundles to improve performance by reducing number of network calls.

    It is not supposed to be instantiated directly by a user. Use :meth:`encord.project.Project.create_bundle()`
    method to initiate bundled operations.

    To execute batch you can either call  :meth:`.execute()` directly, or use a Context Manager.
    """

    def __init__(self, bundle_size: Optional[int] = None) -> None:
        self._bundle_size = bundle_size
        self._operations: Dict[Callable, BundledOperation] = {}

    def __register_operation(
        self,
        payload_type: Type[BundlablePayloadT],
        operation: Callable[..., List[R]],
        result_mapping_predicate: Optional[Callable[[R], str]],
        limit: int,
    ) -> BundledOperation[BundlablePayloadT, R]:
        if operation not in self._operations:
            self._operations[operation] = BundledOperation[BundlablePayloadT, R](
                operation, result_mapping_predicate, limit
            )

        return self._operations[operation]

    def add(
        self,
        operation: Callable[..., List[R]],
        result_mapper: Optional[BundleResultMapper[R]],
        payload: BundlablePayloadT,
        limit: int,
    ) -> None:
        """
        This is an internal method and normally is not supposed to be used externally.

        Adds an operation to a bundle for delayed execution.
        """
        result_mapping_getter = result_mapper.result_mapping_predicate if result_mapper else None
        result_handler = result_mapper.result_handler if result_mapper else None
        self.__register_operation(type(payload), operation, result_mapping_getter, self._bundle_size or limit).append(
            payload, result_handler
        )

    def execute(self) -> None:
        """
        Executes all scheduled operations in bundles and populates results
        """

        for operation in self._operations.values():
            for bundled_payload in operation.get_bundled_payload():
                bundle_result = operation.operation(**asdict(bundled_payload))

                if operation.result_mapper is not None:
                    for br in bundle_result:
                        result_handler = operation.result_handlers.get(operation.result_mapper(br))
                        if result_handler is not None:
                            result_handler(br)
        self._operations = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            log.warning(f"Cancelling operation due to exception: {exc_type.__name__}")
        else:
            self.execute()


def bundled_operation(
    bundle,
    operation,
    payload: BundlablePayloadT,
    result_mapper: Optional[BundleResultMapper] = None,
    limit: int = LABEL_ROW_BUNDLE_DEFAULT_LIMIT,
) -> None:
    assert is_dataclass(payload), "Bundling only works with dataclasses"
    if not bundle:
        result = operation(**asdict(payload))
        if result_mapper:
            assert len(result) == 1, f"Expected a singular response for a singular request, got {len(result)} items!"
            assert result_mapper.result_mapping_predicate(result[0]) == result_mapper.result_handler.predicate
            result_mapper.result_handler.handler(result[0])
    else:
        bundle.add(
            operation=operation,
            result_mapper=result_mapper,
            payload=payload,
            limit=limit,
        )



from enum import IntEnum

from encord.orm.base_dto import BaseDTO


class ProjectUserRole(IntEnum):
    """
    Enumeration for user roles within a project.

    Attributes:
        ADMIN (int): Represents an admin user with value 0.
        ANNOTATOR (int): Represents an annotator user with value 1.
        REVIEWER (int): Represents a reviewer user with value 2.
        ANNOTATOR_REVIEWER (int): Represents a user who is both an annotator and a reviewer with value 3.
        TEAM_MANAGER (int): Represents a team manager user with value 4.
    """

    ADMIN = (0,)
    ANNOTATOR = (1,)
    REVIEWER = (2,)
    ANNOTATOR_REVIEWER = (3,)
    TEAM_MANAGER = 4


class ProjectUser(BaseDTO):
    """
    Data transfer object representing a user within a project.

    Attributes:
        user_email (str): The email address of the user.
        user_role (ProjectUserRole): The role of the user in the project, defined by the ProjectUserRole enumeration.
        project_hash (str): A unique identifier for the project.
    """

    user_email: str
    user_role: ProjectUserRole
    project_hash: str



import uuid
from datetime import datetime
from typing import List, Optional
from uuid import UUID

from encord.orm.base_dto import BaseDTO, Field


class GetCollectionParams(BaseDTO):
    top_level_folder_uuid: Optional[UUID] = Field(default=None, alias="topLevelFolderUuid")
    collection_uuids: Optional[List[UUID]] = Field(default=[], alias="uuids")
    page_token: Optional[str] = Field(default=None, alias="pageToken")
    page_size: Optional[int] = Field(default=None, alias="pageSize")


class CreateCollectionParams(BaseDTO):
    top_level_folder_uuid: Optional[UUID] = Field(default=None, alias="topLevelFolderUuid")


class GetCollectionItemsParams(BaseDTO):
    page_token: Optional[str] = Field(default=None, alias="pageToken")
    page_size: Optional[int] = Field(default=None, alias="pageSize")


class CreateCollectionPayload(BaseDTO):
    name: str
    description: Optional[str] = ""


class UpdateCollectionPayload(BaseDTO):
    name: Optional[str] = None
    description: Optional[str] = None


class Collection(BaseDTO):
    uuid: uuid.UUID
    name: str
    description: Optional[str]
    created_at: Optional[datetime] = Field(default=None, alias="createdAt")
    last_edited_at: Optional[datetime] = Field(default=None, alias="lastEditedAt")


class GetCollectionsResponse(BaseDTO):
    results: List[Collection]


class CollectionBulkItemRequest(BaseDTO):
    item_uuids: List[uuid.UUID]


class CollectionBulkItemResponse(BaseDTO):
    failed_items: List[uuid.UUID]


class CollectionBulkPresetRequest(BaseDTO):
    preset_uuid: uuid.UUID



import uuid
from datetime import datetime
from typing import Generator, Optional, Union
from uuid import UUID

from encord.exceptions import (
    AuthorisationError,
)
from encord.http.v2.api_client import ApiClient
from encord.orm.filter_preset import FilterPreset as OrmFilterPreset
from encord.orm.filter_preset import (
    FilterPresetDefinition,
    GetPresetParams,
    GetPresetsResponse,
)


class FilterPreset:
    """
    Represents preset in Index.
    Preset is a group of filters persisted which can be re-used for faster data curation.
    """

    def __init__(self, client: ApiClient, orm_preset: OrmFilterPreset):
        self._client = client
        self._preset_instance = orm_preset

    @property
    def uuid(self) -> uuid.UUID:
        """
        Get the preset uuid (i.e. the preset ID).

        Returns:
            str: The preset uuid.
        """
        return self._preset_instance.uuid

    @property
    def name(self) -> str:
        """
        Get the preset name

        Returns:
            str: The preset name.
        """
        return self._preset_instance.name

    @property
    def description(self) -> Optional[str]:
        """
        Get the preset description

        Returns:
            Optional[str]: The preset description.
        """
        return self._preset_instance.description

    @property
    def created_at(self) -> Optional[datetime]:
        """
        Get the preset creation timestamp

        Returns:
            Optional[datetime]: The preset creation timestamp.
        """
        return self._preset_instance.created_at

    @property
    def last_updated_at(self) -> Optional[datetime]:
        """
        Get the preset last update timestamp

        Returns:
            Optional[datetime]: The preset last update timestamp.
        """
        return self._preset_instance.last_updated_at

    @staticmethod
    def _get_preset(api_client: ApiClient, preset_uuid: UUID) -> "FilterPreset":
        params = GetPresetParams(uuids=[preset_uuid])
        orm_item = api_client.get(
            "index/presets",
            params=params,
            result_type=GetPresetsResponse,
        )
        if len(orm_item.results) > 0:
            return FilterPreset(api_client, orm_item.results[0])
        raise AuthorisationError("Collection not found")

    @staticmethod
    def _get_presets(
        api_client: ApiClient,
        preset_uuids,
        page_size: Optional[int] = None,
    ) -> "Generator[FilterPreset]":
        params = GetPresetParams(uuids=preset_uuids, pageSize=page_size)
        paged_items = api_client.get_paged_iterator(
            "index/presets",
            params=params,
            result_type=OrmFilterPreset,
        )
        for item in paged_items:
            yield FilterPreset(api_client, item)

    @staticmethod
    def _list_presets(
        api_client: ApiClient, top_level_folder_uuid: Union[UUID, None], page_size: Optional[int] = None
    ) -> "Generator[FilterPreset]":
        params = GetPresetParams(topLevelFolderUuid=top_level_folder_uuid, pageSize=page_size)
        paged_items = api_client.get_paged_iterator(
            "index/presets",
            params=params,
            result_type=OrmFilterPreset,
        )
        for item in paged_items:
            yield FilterPreset(api_client, item)

    @staticmethod
    def _delete_preset(api_client: ApiClient, preset_uuid: UUID) -> None:
        api_client.delete(
            f"index/presets/{preset_uuid}",
            params=None,
            result_type=None,
        )

    def get_filters_json(self) -> FilterPresetDefinition:
        return self._client.get(
            f"index/presets/{self._preset_instance.uuid}",
            params=None,
            result_type=FilterPresetDefinition,
        )



"""
---
title: "Utilities - Ontology Helper"
slug: "sdk-ref-utilities-ont-helper"
hidden: false
metadata:
  title: "Utilities - Ontology Helper"
  description: "Encord SDK Utilities - Ontology Helper."
category: "64e481b57b6027003f20aaa0"
---
"""

from dataclasses import dataclass
from enum import IntEnum


class OntologyUserRole(IntEnum):
    ADMIN = 0
    USER = 1


@dataclass(frozen=True)
class OntologyWithUserRole:
    """
    This is a helper class denoting the relationship between the current user an an ontology
    """

    user_role: int
    user_email: str
    ontology: dict



from dataclasses import dataclass


@dataclass(frozen=True)
class DatasetWithUserRole:
    """
    This is a helper class denoting the relationship between the current user and a project
    """

    user_role: int
    dataset: dict



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import auto
from typing import Dict, List, Optional, Union
from uuid import UUID

from encord.http.v2.api_client import ApiClient
from encord.orm.analytics import CamelStrEnum
from encord.orm.base_dto import BaseDTO, Field
from encord.orm.dataset import DataUnitError, LongPollingStatus

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore[assignment]


class StorageItemType(CamelStrEnum):
    VIDEO = auto()
    IMAGE = auto()
    IMAGE_GROUP = auto()
    IMAGE_SEQUENCE = auto()
    DICOM_FILE = auto()
    DICOM_SERIES = auto()
    AUDIO = auto()


class StorageUserRole(CamelStrEnum):
    USER = auto()
    ADMIN = auto()


class StorageLocationName(CamelStrEnum):
    ENCORD = auto()
    GCP = auto()
    S3 = auto()
    AZURE = auto()
    OPEN_TELEKOM = auto()  # TODO: Remove when enough people have updated their sdk to include S3_COMPATIBLE below
    DIRECT_ACCESS = auto()
    S3_COMPATIBLE = auto()


class PathElement(BaseDTO):
    uuid: UUID
    parent_uuid: Optional[UUID]
    name: str
    synced_dataset_hash: Optional[UUID]


class StorageFolder(BaseDTO):
    uuid: UUID
    parent: Optional[UUID]
    name: str
    description: str
    client_metadata: Optional[str]
    owner: str
    created_at: datetime
    last_edited_at: datetime
    user_role: StorageUserRole
    synced_dataset_hash: Optional[UUID]
    path_to_root: List[PathElement]


class StorageItem(BaseDTO):
    uuid: UUID
    parent: UUID
    item_type: StorageItemType
    name: str
    description: str
    client_metadata: str
    owner: str
    created_at: datetime
    last_edited_at: datetime
    is_tombstone: bool = False
    """This item has been deleted but the link is retained for consistency reasons.
    Mostly for items in the 'cloud linked folders' that are referenced but aren't present after a re-sync"""
    is_placeholder: bool = False
    """This item has been added to the folder but isn't fully processed yet"""
    backed_data_units_count: int
    storage_location: StorageLocationName
    integration_hash: Optional[UUID]
    url: Optional[str]
    signed_url: Optional[str]
    file_size: Optional[int]
    mime_type: Optional[str]
    duration: Optional[float]
    fps: Optional[float]
    height: Optional[int]
    width: Optional[int]
    dicom_instance_uid: Optional[str]
    dicom_study_uid: Optional[str]
    dicom_series_uid: Optional[str]
    frame_count: Optional[int]
    audio_sample_rate: Optional[int]
    audio_bit_depth: Optional[int]
    audio_codec: Optional[str]
    audio_num_channels: Optional[int]


class StorageItemInaccessible(BaseDTO):
    uuid: UUID


class CreateStorageFolderPayload(BaseDTO):
    name: str
    description: Optional[str]
    parent: Optional[UUID]
    client_metadata: Optional[str]


class UploadSignedUrlsPayload(BaseDTO):
    item_type: StorageItemType
    count: int
    frames_subfolder_name: Optional[str]


class UploadSignedUrl(BaseDTO):
    item_uuid: UUID
    object_key: str
    signed_url: str


class StorageItemWithName(BaseDTO):
    item_uuid: UUID
    name: str


class UploadLongPollingState(BaseDTO):
    """
    Response of the upload job's long polling request.

    Note: An upload job consists of job units, where job unit could be
    either a video, image group, dicom series, or a single image.
    """

    status: LongPollingStatus
    """Status of the upload job. Documented in detail in :meth:`encord.orm.dataset.LongPollingStatus`"""

    items_with_names: List[StorageItemWithName]
    """Information about data which was added to the folder."""

    errors: List[str]
    """Stringified list of exceptions."""

    units_pending_count: int
    """Number of upload job units that have pending status."""

    units_done_count: int
    """Number of upload job units that have done status."""

    units_error_count: int
    """Number of upload job units that have error status."""

    unit_errors: List[DataUnitError]
    """Structured list of per-item upload errors. See :class:`DataUnitError` for more details."""

    file_name: Optional[str] = None
    """Name of the JSON or CSV file that contained the list of URLs to ingest form the cloud bucket. Optional."""


class CustomerProvidedVideoMetadata(BaseDTO):
    """
    Media metadata for a video file; if provided, Encord service will skip frame synchronisation checks
    and will use the values specified here to render the video in the label editor.
    """

    fps: float
    """Frame rate of the video in frames per second."""
    duration: float
    """Video duration in (float) seconds."""
    width: int
    """Width of the video in pixels."""
    height: int
    """Height of the video in pixels."""
    file_size: int
    """Size of the video file in bytes."""
    mime_type: str
    """MIME type of the video file (e.g. `video/mp4` or `video/webm`)."""


class CustomerProvidedAudioMetadata(BaseDTO):
    """
    Media metadata for an audio file; if provided, Encord service will use the values here instead of scanning the files
    """

    duration: float
    """Audio duration in (float) seconds."""
    file_size: int
    """Size of the audio file in bytes."""
    mime_type: str
    """MIME type of the audio file (for example: `audio/mpeg` or `audio/wav`)."""
    sample_rate: int
    """Sample rate (int) in Hz."""
    bit_depth: int
    """Size of each sample (int) in bits."""
    codec: str
    """Codec (for example: mp3, pcm)."""
    num_channels: int
    """Number of channels"""


class DataUploadImage(BaseDTO):
    object_url: str
    title: Optional[str] = None
    client_metadata: Dict = Field(default_factory=dict)
    external_file_type: Literal["IMAGE"] = "IMAGE"

    placeholder_item_uuid: Optional[UUID] = None


class DataUploadVideo(BaseDTO):
    object_url: str
    title: Optional[str] = None
    client_metadata: Dict = Field(default_factory=dict)

    external_file_type: Literal["VIDEO"] = "VIDEO"
    video_metadata: Optional[CustomerProvidedVideoMetadata] = None

    placeholder_item_uuid: Optional[UUID] = None


class DataUploadImageGroupImage(BaseDTO):
    url: str
    title: Optional[str]

    placeholder_item_uuid: Optional[UUID] = None


class DataUploadImageGroup(BaseDTO):
    images: List[DataUploadImageGroupImage]
    title: Optional[str]
    client_metadata: Dict = Field(default_factory=dict)
    create_video: bool = False

    external_file_type: Literal["IMG_GROUP"] = "IMG_GROUP"
    cluster_by_resolution: bool = False


class DataUploadImageGroupFromItems(BaseDTO):
    image_items: List[UUID]
    title: Optional[str]
    client_metadata: Dict = Field(default_factory=dict)
    create_video: bool
    video_url_prefix: Optional[str] = None

    external_file_type: Literal["IMG_GROUP_FROM_ITEMS"] = "IMG_GROUP_FROM_ITEMS"
    cluster_by_resolution: bool = False


class DataUploadDicomSeriesDicomFile(BaseDTO):
    url: str
    title: Optional[str]

    placeholder_item_uuid: Optional[UUID] = None


class DataUploadDicomSeries(BaseDTO):
    dicom_files: List[DataUploadDicomSeriesDicomFile]
    title: Optional[str]
    client_metadata: Dict = Field(default_factory=dict)

    external_file_type: Literal["DICOM"] = "DICOM"


class DataUploadAudio(BaseDTO):
    object_url: str
    title: Optional[str] = None
    client_metadata: Dict = Field(default_factory=dict)

    audio_metadata: Optional[CustomerProvidedAudioMetadata] = None
    external_file_type: Literal["AUDIO"] = "AUDIO"

    placeholder_item_uuid: Optional[UUID] = None


class DataUploadItems(BaseDTO):
    videos: List[DataUploadVideo] = Field(default_factory=list)
    image_groups: List[DataUploadImageGroup] = Field(default_factory=list)
    dicom_series: List[DataUploadDicomSeries] = Field(default_factory=list)
    images: List[DataUploadImage] = Field(default_factory=list)
    image_groups_from_items: List[DataUploadImageGroupFromItems] = Field(default_factory=list)
    audio: List[DataUploadAudio] = Field(default_factory=list)
    skip_duplicate_urls: bool = False


class PostUploadJobParams(BaseDTO):
    data_items: Optional[DataUploadItems] = None
    external_files: Optional[dict] = None
    integration_hash: Optional[UUID] = None
    ignore_errors: bool = False
    file_name: Optional[str] = None


class GetUploadJobParams(BaseDTO):
    timeout_seconds: int = 60


class FoldersSortBy(CamelStrEnum):
    NAME = auto()
    CREATED_AT = auto()


class ListItemsParams(BaseDTO):
    search: Optional[str]
    is_recursive: Optional[bool] = False
    is_in_dataset: Optional[bool]
    item_types: List[StorageItemType]
    include_org_access: Optional[bool] = None
    order: FoldersSortBy
    desc: bool
    page_token: Optional[str]
    page_size: int
    sign_urls: bool


class ListFoldersParams(BaseDTO):
    search: Optional[str] = None
    is_recursive: Optional[bool] = False
    dataset_synced: Optional[bool] = None
    include_org_access: Optional[bool] = None
    order: FoldersSortBy = FoldersSortBy.NAME
    desc: bool = False
    page_token: Optional[str] = None
    page_size: int = 100


class PatchItemPayload(BaseDTO):
    name: Optional[str] = None
    description: Optional[str] = None
    client_metadata: Optional[dict] = None


class PatchFolderPayload(BaseDTO):
    name: Optional[str] = None
    description: Optional[str] = None
    client_metadata: Optional[dict] = None


class PatchFoldersBulkPayload(BaseDTO):
    folder_patches: Dict[str, PatchFolderPayload]


class StorageFolderSummary(BaseDTO):
    files: int
    folders: int
    videos: int
    images: int
    image_groups: int
    image_sequences: int
    dicom_files: int
    dicom_series: int
    tombstones: int
    upload_jobs: int
    total_size: float


class ItemShortInfo(BaseDTO):
    uuid: UUID
    name: str
    parent_uuid: UUID
    parent_name: str
    item_type: StorageItemType


class DatasetShortInfo(BaseDTO):
    dataset_hash: str
    backing_folder_uuid: Optional[UUID]
    title: str
    data_hashes: List[UUID]
    data_units_created_at: datetime


class StorageItemSummary(BaseDTO):
    """
    A summary of item usage in the system
    """

    frame_in_groups: int
    """A number of group items (DICOM_SERIES, IMAGE_GROUP, IMAGE_SEQUENCE) that contain this item"""
    accessible_group_items: List[ItemShortInfo]
    """List of group items that contain this item (only those that the user has access to,
    so the length of this list can be less than `frame_in_groups`)"""
    used_in_datasets: int
    """A number of datasets that contain this item as a `DataRow`"""
    accessible_datasets: List[DatasetShortInfo]
    """List of datasets that contain this item as a `DataRow` (only those that the user has access to, so
    the length of this list can be less than `used_in_datasets`)"""


class DeleteItemsParams(BaseDTO):
    remove_unused_frames: bool


class DeleteItemsPayload(BaseDTO):
    child_uuids: List[UUID]
    remove_unused_frames: bool


class DeleteItemsResponse(BaseDTO):
    removed_items_count: int
    removed_folders_count: int


class MoveItemsPayload(BaseDTO):
    item_uuids: List[UUID]
    new_parent_uuid: UUID
    allow_synced_dataset_move: bool = False


class MoveFoldersPayload(BaseDTO):
    folder_uuids: List[UUID]
    new_parent_uuid: Optional[UUID]


class GetItemParams(BaseDTO):
    sign_url: bool


class GetChildItemsParams(BaseDTO):
    sign_urls: bool = False


class GetItemsBulkPayload(BaseDTO):
    item_uuids: List[UUID]
    sign_urls: bool = False


class PatchItemsBulkPayload(BaseDTO):
    item_patches: Dict[str, PatchItemPayload]


@dataclass
class BundledPatchItemPayload:
    item_patches: Dict[str, PatchItemPayload]

    def add(self, other: BundledPatchItemPayload) -> BundledPatchItemPayload:
        self.item_patches.update(other.item_patches)
        return self


@dataclass
class BundledPatchFolderPayload:
    folder_patches: Dict[str, PatchFolderPayload]

    def add(self, other: BundledPatchFolderPayload) -> BundledPatchFolderPayload:
        self.folder_patches.update(other.folder_patches)
        return self


class ReencodeVideoItemsRequest(BaseDTO):
    storage_items: List[UUID]
    process_title: str
    force_full_reencoding: bool


class JobStatus(CamelStrEnum):
    SUBMITTED = auto()
    DONE = auto()
    ERROR = auto()


class ReencodeVideoItemsResponse(BaseDTO):
    status: JobStatus
    result: Optional[Union[list, dict]]


class StorageItemsMigratePayload(BaseDTO):
    urls_map: Optional[Dict[str, Optional[str]]]
    items_map: Optional[Dict[UUID, Optional[str]]]
    from_integration_hash: Optional[UUID] = None
    to_integration_hash: Optional[UUID] = None
    validate_access: bool = False
    skip_missing: bool = False



ENCORD_CONTACT_SUPPORT_EMAIL = "support@encord.com"


def _get_dict_without_none_keys(d: dict) -> dict:
    """Remove keys with value None from a dictionary. Does not change the dict in place."""
    return {k: v for k, v in d.items() if v is not None}



from __future__ import annotations

from enum import Enum
from typing import Optional, Type, TypeVar, cast

T = TypeVar("T", bound="StringEnum")


class StringEnum(str, Enum):
    """
    Use this enum class if you need the helper that creates the enum instance from a string.
    """

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls: Type[T], value: str) -> Optional[T]:
        # pylint: disable-next=no-member
        return cast(T, cls._value2member_map_.get(value))



"""
---
title: "Objects - Coordinates"
slug: "sdk-ref-objects-coordinates"
hidden: false
metadata:
  title: "Objects - Coordinates"
  description: "Encord SDK Objects - Coordinates."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Flag, auto
from typing import Any, Dict, List, Optional, Type, Union

from encord.exceptions import LabelRowError
from encord.objects.bitmask import BitmaskCoordinates
from encord.objects.common import Shape
from encord.orm.base_dto import BaseDTO


@dataclass(frozen=True)
class BoundingBoxCoordinates:
    """
    Represents bounding box coordinates, where all values are percentages relative to the total image size.

    Attributes:
        height (float): The height of the bounding box.
        width (float): The width of the bounding box.
        top_left_x (float): The x-coordinate of the top-left corner.
        top_left_y (float): The y-coordinate of the top-left corner.
    """

    height: float
    width: float
    top_left_x: float
    top_left_y: float

    @staticmethod
    def from_dict(d: dict) -> BoundingBoxCoordinates:
        """
        Create a BoundingBoxCoordinates instance from a dictionary.

        Args:
            d (dict): A dictionary containing bounding box information.

        Returns:
            BoundingBoxCoordinates: An instance of BoundingBoxCoordinates.
        """
        bounding_box_dict = d["boundingBox"]
        return BoundingBoxCoordinates(
            height=bounding_box_dict["h"],
            width=bounding_box_dict["w"],
            top_left_x=bounding_box_dict["x"],
            top_left_y=bounding_box_dict["y"],
        )

    def to_dict(self) -> dict:
        """
        Convert the BoundingBoxCoordinates instance to a dictionary.

        Returns:
            dict: A dictionary representation of the bounding box coordinates.
        """
        return {
            "h": self.height,
            "w": self.width,
            "x": self.top_left_x,
            "y": self.top_left_y,
        }


@dataclass(frozen=True)
class RotatableBoundingBoxCoordinates:
    """
    Represents rotatable bounding box coordinates, where all values are percentages relative to the total image size.

    Attributes:
        height (float): The height of the bounding box.
        width (float): The width of the bounding box.
        top_left_x (float): The x-coordinate of the top-left corner.
        top_left_y (float): The y-coordinate of the top-left corner.
        theta (float): The angle of rotation originating at the center of the box.
    """

    height: float
    width: float
    top_left_x: float
    top_left_y: float
    theta: float  # angle of rotation originating at center of box

    @staticmethod
    def from_dict(d: dict) -> RotatableBoundingBoxCoordinates:
        """
        Create a RotatableBoundingBoxCoordinates instance from a dictionary.

        Args:
            d (dict): A dictionary containing rotatable bounding box information.

        Returns:
            RotatableBoundingBoxCoordinates: An instance of RotatableBoundingBoxCoordinates.
        """
        rotatable_bounding_box_dict = d["rotatableBoundingBox"]
        return RotatableBoundingBoxCoordinates(
            height=rotatable_bounding_box_dict["h"],
            width=rotatable_bounding_box_dict["w"],
            top_left_x=rotatable_bounding_box_dict["x"],
            top_left_y=rotatable_bounding_box_dict["y"],
            theta=rotatable_bounding_box_dict["theta"],
        )

    def to_dict(self) -> dict:
        """
        Convert the RotatableBoundingBoxCoordinates instance to a dictionary.

        Returns:
            dict: A dictionary representation of the rotatable bounding box coordinates.
        """
        return {
            "h": self.height,
            "w": self.width,
            "x": self.top_left_x,
            "y": self.top_left_y,
            "theta": self.theta,
        }


@dataclass(frozen=True)
class PointCoordinate:
    """
    Represents a point coordinate, where all coordinates are a percentage relative to the total image size.

    Attributes:
        x (float): The x-coordinate of the point.
        y (float): The y-coordinate of the point.
    """

    x: float
    y: float

    @staticmethod
    def from_dict(d: dict) -> PointCoordinate:
        """
        Create a PointCoordinate instance from a dictionary.

        Args:
            d (dict): A dictionary containing point coordinate information.

        Returns:
            PointCoordinate: An instance of PointCoordinate.
        """
        first_item = d["point"]["0"]
        return PointCoordinate(
            x=first_item["x"],
            y=first_item["y"],
        )

    def to_dict(self) -> dict:
        """
        Convert the PointCoordinate instance to a dictionary.

        Returns:
            dict: A dictionary representation of the point coordinate.
        """
        return {"0": {"x": self.x, "y": self.y}}


@dataclass(frozen=True)
class PolygonCoordinates:
    """
    Represents polygon coordinates as a list of point coordinates.

    Attributes:
        values (List[PointCoordinate]): A list of PointCoordinate objects defining the polygon.
    """

    values: List[PointCoordinate]

    @staticmethod
    def from_dict(d: dict) -> PolygonCoordinates:
        """
        Create a PolygonCoordinates instance from a dictionary.

        Args:
            d (dict): A dictionary containing polygon coordinates information.

        Returns:
            PolygonCoordinates: An instance of PolygonCoordinates.
        """
        polygon_dict = d["polygon"]
        values: List[PointCoordinate] = []

        if isinstance(polygon_dict, dict):
            sorted_dict_value_tuples = sorted((int(key), value) for key, value in polygon_dict.items())
            sorted_dict_values = [item[1] for item in sorted_dict_value_tuples]
        elif isinstance(polygon_dict, list):
            sorted_dict_values = list(polygon_dict)
        else:
            raise LabelRowError(f"Invalid format for polygon coordinates: {polygon_dict}")

        for value in sorted_dict_values:
            point_coordinate = PointCoordinate(
                x=value["x"],
                y=value["y"],
            )
            values.append(point_coordinate)

        return PolygonCoordinates(values=values)

    def to_dict(self) -> dict:
        """
        Convert the PolygonCoordinates instance to a dictionary.

        Returns:
            dict: A dictionary representation of the polygon coordinates.
        """
        return {str(idx): {"x": value.x, "y": value.y} for idx, value in enumerate(self.values)}


@dataclass(frozen=True)
class PolylineCoordinates:
    """
    Represents polyline coordinates as a list of point coordinates.

    Attributes:
        values (List[PointCoordinate]): A list of PointCoordinate objects defining the polyline.
    """

    values: List[PointCoordinate]

    @staticmethod
    def from_dict(d: dict) -> PolylineCoordinates:
        """
        Create a PolylineCoordinates instance from a dictionary.

        Args:
            d (dict): A dictionary containing polyline coordinates information.

        Returns:
            PolylineCoordinates: An instance of PolylineCoordinates.
        """
        polyline = d["polyline"]
        values: List[PointCoordinate] = []

        if isinstance(polyline, dict):
            sorted_dict_value_tuples = sorted((int(key), value) for key, value in polyline.items())
            sorted_dict_values = [item[1] for item in sorted_dict_value_tuples]
        elif isinstance(polyline, list):
            sorted_dict_values = list(polyline)
        else:
            raise LabelRowError(f"Invalid format for polyline coordinates: {polyline}")

        for value in sorted_dict_values:
            point_coordinate = PointCoordinate(
                x=value["x"],
                y=value["y"],
            )
            values.append(point_coordinate)

        return PolylineCoordinates(values=values)

    def to_dict(self) -> dict:
        """
        Convert the PolylineCoordinates instance to a dictionary.

        Returns:
            dict: A dictionary representation of the polyline coordinates.
        """
        return {str(idx): {"x": value.x, "y": value.y} for idx, value in enumerate(self.values)}


class Visibility(Flag):
    """
    An enumeration to represent the visibility state of an item.

    Attributes:
        VISIBLE: The item is visible within the frame.
        INVISIBLE: The item is outside the frame and thus invisible.
        OCCLUDED: The item is within the frame but occluded by something else.
    """

    VISIBLE = auto()
    INVISIBLE = auto()
    OCCLUDED = auto()


class SkeletonCoordinate(BaseDTO):
    """
    Represents a coordinate for a skeleton structure in an image.

    Attributes:
        x (float): The x-coordinate of the skeleton point.
        y (float): The y-coordinate of the skeleton point.
        name (str): The name of the skeleton point.
        color (Optional[str]): The color associated with the skeleton point.
        feature_hash (Optional[str]): A unique hash for the feature.
        value (Optional[str]): An optional value associated with the skeleton point.
        visibility (Optional[Visibility]): The visibility state of the skeleton point.
    """

    x: float
    y: float
    name: str
    color: Optional[str] = None
    feature_hash: Optional[str] = None
    value: Optional[str] = None

    visibility: Optional[Visibility] = None


class SkeletonCoordinates(BaseDTO):
    """
    Represents a collection of skeleton coordinates.

    Attributes:
        values (List[SkeletonCoordinate]): A list of SkeletonCoordinate objects.
        name (str): The name of the skeleton structure.
    """

    values: List[SkeletonCoordinate]
    name: str

    def to_dict(self, by_alias=True, exclude_none=True) -> Dict[str, Any]:
        """
        Convert the SkeletonCoordinates instance to a dictionary.

        Args:
            by_alias (bool): Whether to use alias for the field names.
            exclude_none (bool): Whether to exclude fields with None values.

        Returns:
            Dict[str, Any]: A dictionary representation of the skeleton coordinates.
        """
        return {str(i): x.to_dict() for i, x in enumerate(self.values)}


Coordinates = Union[
    BoundingBoxCoordinates,
    RotatableBoundingBoxCoordinates,
    PointCoordinate,
    PolygonCoordinates,
    PolylineCoordinates,
    SkeletonCoordinates,
    BitmaskCoordinates,
]
ACCEPTABLE_COORDINATES_FOR_ONTOLOGY_ITEMS: Dict[Shape, Type[Coordinates]] = {
    Shape.BOUNDING_BOX: BoundingBoxCoordinates,
    Shape.ROTATABLE_BOUNDING_BOX: RotatableBoundingBoxCoordinates,
    Shape.POINT: PointCoordinate,
    Shape.POLYGON: PolygonCoordinates,
    Shape.POLYLINE: PolylineCoordinates,
    Shape.SKELETON: SkeletonCoordinates,
    Shape.BITMASK: BitmaskCoordinates,
}



"""
---
title: "CloudUpload"
slug: "sdk-ref-cloudupload"
hidden: false
metadata:
  title: "CloudUpload"
  description: "Encord SDK CloudUpload class."
category: "64e481b57b6027003f20aaa0"
---
"""

import logging
import mimetypes
import os.path
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Optional, Type, Union
from uuid import UUID

from tqdm import tqdm

from encord.configs import BaseConfig
from encord.exceptions import CloudUploadError
from encord.http.querier import Querier, create_new_session
from encord.orm.dataset import (
    Audio,
    DicomSeries,
    Images,
    SignedAudioURL,
    SignedDicomsURL,
    SignedDicomURL,
    SignedImagesURL,
    SignedImageURL,
    SignedVideoURL,
    Video,
)

PROGRESS_BAR_FILE_FACTOR = 100
CACHE_DURATION_IN_SECONDS = 24 * 60 * 60  # 1 day

logger = logging.getLogger(__name__)


@dataclass
class CloudUploadSettings:
    """
    The settings for uploading data into the GCP cloud storage. These apply for each individual upload. These settings
    will overwrite the :meth:`encord.http.constants.RequestsSettings` which is set during
    :class:`encord.EncordUserClient` creation.
    """

    max_retries: Optional[int] = None
    """Number of allowed retries when uploading"""
    backoff_factor: Optional[float] = None
    """With each retry, there will be a sleep of backoff_factor * (2 ** (retry_number - 1) )"""
    allow_failures: bool = False
    """
    If failures are allowed, the upload will continue even if some items were not successfully uploaded even
    after retries. For example, upon creation of a large image group, you might want to create the image group
    even if a few images were not successfully uploaded. The unsuccessfully uploaded images will then be logged.
    """


def _get_content_type(
    orm_class: Union[Type[Images], Type[Video], Type[DicomSeries], Type[Audio]], file_path: str
) -> Optional[str]:
    if orm_class == Images:
        return mimetypes.guess_type(file_path)[0]
    elif orm_class == Video or orm_class == Audio:
        return "application/octet-stream"
    elif orm_class == DicomSeries:
        return "application/dicom"
    else:
        raise ValueError(f"Unsupported type `{orm_class}`")


def upload_to_signed_url_list(
    file_paths: Iterable[Union[str, Path]],
    config: BaseConfig,
    querier: Querier,
    orm_class: Union[Type[Images], Type[Video], Type[DicomSeries], Type[Audio]],
    cloud_upload_settings: CloudUploadSettings,
) -> List[Union[SignedVideoURL, SignedImageURL, SignedDicomURL, SignedAudioURL]]:
    """Upload files and return the upload returns in the same order as the file paths supplied."""
    failed_uploads = []
    successful_uploads = []
    for file_path in tqdm(file_paths):
        file_path = str(file_path)
        content_type = _get_content_type(orm_class, file_path)
        file_name = os.path.basename(file_path)
        signed_url = _get_signed_url(file_name, orm_class, querier)
        assert signed_url.get("title", "") == file_name, "Ordering issue"

        try:
            if cloud_upload_settings.max_retries is not None:
                max_retries = cloud_upload_settings.max_retries
            else:
                max_retries = config.requests_settings.max_retries

            if cloud_upload_settings.backoff_factor is not None:
                backoff_factor = cloud_upload_settings.backoff_factor
            else:
                backoff_factor = config.requests_settings.backoff_factor

            _upload_single_file(
                file_path, signed_url, content_type, max_retries=max_retries, backoff_factor=backoff_factor
            )
            successful_uploads.append(signed_url)
        except CloudUploadError as e:
            if cloud_upload_settings.allow_failures:
                failed_uploads.append(file_path)
            else:
                raise e

    if failed_uploads:
        logger.warning("The upload was incomplete for the following items: %s", failed_uploads)

    return successful_uploads


def upload_video_to_encord(
    signed_url: Union[SignedVideoURL, SignedImageURL, SignedDicomURL, SignedAudioURL],
    video_title: Optional[str],
    folder_uuid: Optional[UUID],
    querier: Querier,
) -> Video:
    payload = {
        "signed_url": signed_url["signed_url"],
        "data_hash": signed_url["data_hash"],
        "title": signed_url["title"],
        "file_link": signed_url["file_link"],
        "video_title": video_title,
    }
    if folder_uuid is not None:
        payload["folder_uuid"] = str(folder_uuid)

    return querier.basic_put(
        Video,
        uid=signed_url.get("data_hash"),
        payload=payload,
        enable_logging=False,
    )


def upload_images_to_encord(
    signed_urls: List[Union[SignedVideoURL, SignedImageURL, SignedDicomURL, SignedAudioURL]], querier: Querier
) -> Images:
    return querier.basic_put(Images, uid=None, payload=signed_urls, enable_logging=False)


def _get_signed_url(
    file_name: str, orm_class: Union[Type[Images], Type[Video], Type[DicomSeries], Type[Audio]], querier: Querier
) -> Union[SignedVideoURL, SignedImageURL, SignedDicomURL, SignedAudioURL]:
    if orm_class == Video:
        return querier.basic_getter(SignedVideoURL, uid=file_name)
    elif orm_class == Images:
        return querier.basic_getter(SignedImagesURL, uid=[file_name])[0]
    elif orm_class == Audio:
        return querier.basic_getter(SignedAudioURL, uid=file_name)
    elif orm_class == DicomSeries:
        return querier.basic_getter(SignedDicomsURL, uid=[file_name])[0]
    raise ValueError(f"Unsupported type `{orm_class}`")


def _upload_single_file(
    file_path: Union[str, Path],
    signed_url: dict,
    content_type: Optional[str],
    *,
    max_retries: int,
    backoff_factor: float,
    cache_max_age: int = CACHE_DURATION_IN_SECONDS,
) -> None:
    with create_new_session(
        max_retries=max_retries, backoff_factor=backoff_factor, connect_retries=max_retries
    ) as session:
        url = signed_url["signed_url"]

        with open(file_path, "rb") as f:
            res_upload = session.put(
                url, data=f, headers={"Content-Type": content_type, "Cache-Control": f"max-age={cache_max_age}"}
            )

            if res_upload.status_code != 200:
                status_code = res_upload.status_code
                headers = res_upload.headers
                res_text = res_upload.text
                error_string = str(
                    f"Error uploading file '{signed_url.get('title', '')}' to signed url: "
                    f"'{signed_url.get('signed_url')}'.\n"
                    f"Response data:\n\tstatus code: '{status_code}'\n\theaders: '{headers}'\n\tcontent: '{res_text}'",
                )

                logger.error(error_string)
                raise CloudUploadError(error_string)



"""``encord.client`` provides a simple Python client that allows you
to query project resources through the Encord API.

Here is a simple example for instantiating the client for a project
and obtaining project info:

.. test_blurb2.py code::

    from encord.client import EncordClient

    client = EncordClient.initialize('YourProjectID', 'YourAPIKey')
    client.get_project()

    Returns:
        Project: A project record instance. See Project ORM for details.

"""

from __future__ import annotations

import base64
import dataclasses
import json
import logging
import os.path
import time
import typing
import uuid
from datetime import datetime
from math import ceil
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Union, cast
from uuid import UUID

import requests

import encord.exceptions
from encord.common.deprecated import deprecated
from encord.configs import ENCORD_DOMAIN, ApiKeyConfig, BearerConfig, Config, EncordConfig, SshConfig
from encord.constants.enums import DataType
from encord.constants.model import AutomationModels, Device
from encord.constants.string_constants import (
    FITTED_BOUNDING_BOX,
    INTERPOLATION,
    TYPE_DATASET,
    TYPE_PROJECT,
)
from encord.exceptions import EncordException
from encord.http.constants import DEFAULT_REQUESTS_SETTINGS, RequestsSettings
from encord.http.querier import Querier
from encord.http.utils import (
    CloudUploadSettings,
    upload_images_to_encord,
    upload_to_signed_url_list,
    upload_video_to_encord,
)
from encord.http.v2.api_client import ApiClient
from encord.http.v2.payloads import Page
from encord.orm.analytics import (
    CollaboratorTimer,
    CollaboratorTimerParams,
)
from encord.orm.api_key import ApiKeyMeta
from encord.orm.bearer_request import BearerTokenResponse
from encord.orm.cloud_integration import CloudIntegration
from encord.orm.dataset import (
    DEFAULT_DATASET_ACCESS_SETTINGS,
    AddPrivateDataResponse,
    Audio,
    DataLinkDuplicatesBehavior,
    DataRow,
    DataRows,
    DatasetAccessSettings,
    DatasetData,
    DatasetDataLongPolling,
    DatasetLinkItems,
    DatasetUser,
    DatasetUserRole,
    DatasetUsers,
    DicomSeries,
    Image,
    ImageGroup,
    ImageGroupOCR,
    Images,
    LongPollingStatus,
    ReEncodeVideoTask,
    SingleImage,
    Video,
)
from encord.orm.dataset import Dataset as OrmDataset
from encord.orm.group import (
    AddDatasetGroupsPayload,
    AddProjectGroupsPayload,
    DatasetGroup,
    ProjectGroup,
    RemoveGroupsParams,
)
from encord.orm.label_log import LabelLog, LabelLogParams
from encord.orm.label_row import (
    AnnotationTaskStatus,
    LabelRow,
    LabelRowMetadata,
    LabelStatus,
    LabelValidationState,
    Review,
    ShadowDataState,
)
from encord.orm.labeling_algorithm import (
    BoundingBoxFittingParams,
    LabelingAlgorithm,
    ObjectInterpolationParams,
)
from encord.orm.model import (
    Model,
    ModelConfiguration,
    ModelInferenceParams,
    ModelOperations,
    ModelRow,
    ModelTrainingWeights,
    PublicModelTrainGetResultLongPollingStatus,
    PublicModelTrainGetResultParams,
    PublicModelTrainGetResultResponse,
    PublicModelTrainStartPayload,
    TrainingMetadata,
)
from encord.orm.project import (
    CopyDatasetAction,
    CopyDatasetOptions,
    CopyLabelsOptions,
    CopyProjectPayload,
    ProjectCopy,
    ProjectCopyOptions,
    ProjectDataset,
    ProjectUsers,
    TaskPriorityParams,
)
from encord.orm.project import Project as OrmProject
from encord.orm.project import ProjectDTO as ProjectOrmV2
from encord.orm.workflow import (
    LabelWorkflowGraphNode,
    LabelWorkflowGraphNodePayload,
    WorkflowAction,
)
from encord.project_ontology.classification_type import ClassificationType
from encord.project_ontology.object_type import ObjectShape
from encord.project_ontology.ontology import Ontology as LegacyOntology
from encord.utilities.client_utilities import optional_datetime_to_iso_str, optional_set_to_list
from encord.utilities.project_user import ProjectUser, ProjectUserRole

LONG_POLLING_RESPONSE_RETRY_N = 3
LONG_POLLING_SLEEP_ON_FAILURE_SECONDS = 10
LONG_POLLING_MAX_REQUEST_TIME_SECONDS = 60

logger = logging.getLogger(__name__)


class EncordClient:
    """
    Encord client. Allows you to query db items associated
    with a project (e.g. label rows, datasets).
    """

    def __init__(self, querier: Querier, config: Config, api_client: Optional[ApiClient] = None):
        self._querier = querier
        self._config = config
        self._api_client = api_client

    def _get_api_client(self) -> ApiClient:
        if not (isinstance(self._config, (SshConfig, BearerConfig))):
            raise EncordException(
                "This functionality requires private SSH key authentication. API keys are not supported."
            )

        if not self._api_client:
            raise RuntimeError("ApiClient should exist when authenticated with SSH key.")

        return self._api_client

    @staticmethod
    def initialise(
        resource_id: Optional[str] = None,
        api_key: Optional[str] = None,
        domain: str = ENCORD_DOMAIN,
        requests_settings: RequestsSettings = DEFAULT_REQUESTS_SETTINGS,
    ) -> Union[EncordClientProject, EncordClientDataset]:
        """
        Create and initialize a Encord client from a resource EntityId and API key.

        Args:
            resource_id: either of the following

                * A <project_hash>.
                  If ``None``, uses the ``ENCORD_PROJECT_ID`` environment variable.
                  The ``CORD_PROJECT_ID`` environment variable is supported for backwards compatibility.

                * A <dataset_hash>.
                  If ``None``, uses the ``ENCORD_DATASET_ID`` environment variable.
                  The ``CORD_DATASET_ID`` environment variable is supported for backwards compatibility.

            api_key: An API key.
                     If None, uses the ``ENCORD_API_KEY`` environment variable.
                     The ``CORD_API_KEY`` environment variable is supported for backwards compatibility.
            domain: The encord api-server domain.
                If None, the :obj:`encord.configs.ENCORD_DOMAIN` is used
            requests_settings: The RequestsSettings from this config

        Returns:
            EncordClient: A Encord client instance.
        """
        config = EncordConfig(resource_id, api_key, domain=domain, requests_settings=requests_settings)
        return EncordClient.initialise_with_config(config)

    @staticmethod
    def initialise_with_config(config: ApiKeyConfig) -> Union[EncordClientProject, EncordClientDataset]:
        """
        Create and initialize a Encord client from a Encord config instance.

        Args:
            config: A Encord config instance.

        Returns:
            EncordClient: A Encord client instance.
        """
        querier = Querier(config, resource_id=config.resource_id)
        key_type = querier.basic_getter(ApiKeyMeta)

        if key_type.resource_type == TYPE_PROJECT:
            logger.info("Initialising Encord client for project using key: %s", key_type.title)
            return EncordClientProject(querier, config)

        elif key_type.resource_type == TYPE_DATASET:
            logger.info("Initialising Encord client for dataset using key: %s", key_type.title)
            return EncordClientDataset(querier, config)

        else:
            raise encord.exceptions.InitialisationError(
                message=f"API key [{config.api_key}] is not associated with a project or dataset"
            )

    def get_cloud_integrations(self) -> List[CloudIntegration]:
        return self._querier.get_multiple(CloudIntegration)

    def get_bearer_token(self) -> BearerTokenResponse:
        return self._get_api_client().get("user/bearer_token", None, result_type=BearerTokenResponse)


class EncordClientDataset(EncordClient):
    """
    DEPRECATED - prefer using the :class:`encord.dataset.Dataset` instead
    """

    def __init__(
        self,
        querier: Querier,
        config: Config,
        dataset_access_settings: DatasetAccessSettings = DEFAULT_DATASET_ACCESS_SETTINGS,
        api_client: Optional[ApiClient] = None,
    ):
        super().__init__(querier, config, api_client)
        self._dataset_access_settings = dataset_access_settings

    @staticmethod
    def initialise(
        resource_id: Optional[str] = None,
        api_key: Optional[str] = None,
        domain: str = ENCORD_DOMAIN,
        requests_settings: RequestsSettings = DEFAULT_REQUESTS_SETTINGS,
        dataset_access_settings: DatasetAccessSettings = DEFAULT_DATASET_ACCESS_SETTINGS,
    ) -> EncordClientDataset:
        """
        Create and initialize a Encord client from a resource EntityId and API key.

        Args:
            resource_id: either of the following

                * A <project_hash>.
                  If ``None``, uses the ``ENCORD_PROJECT_ID`` environment variable.
                  The ``CORD_PROJECT_ID`` environment variable is supported for backwards compatibility.

                * A <dataset_hash>.
                  If ``None``, uses the ``ENCORD_DATASET_ID`` environment variable.
                  The ``CORD_DATASET_ID`` environment variable is supported for backwards compatibility.

            api_key: An API key.
                     If None, uses the ``ENCORD_API_KEY`` environment variable.
                     The ``CORD_API_KEY`` environment variable is supported for backwards compatibility.
            domain: The encord api-server domain.
                If None, the :obj:`encord.configs.ENCORD_DOMAIN` is used
            requests_settings: The RequestsSettings from this config
            dataset_access_settings: Change the default :class:`encord.orm.dataset.DatasetAccessSettings`.

        Returns:
            EncordClientDataset: A Encord client dataset instance.
        """
        config = EncordConfig(resource_id, api_key, domain=domain, requests_settings=requests_settings)
        return EncordClientDataset.initialise_with_config(config, dataset_access_settings=dataset_access_settings)

    @staticmethod
    def initialise_with_config(
        config: ApiKeyConfig, dataset_access_settings: DatasetAccessSettings = DEFAULT_DATASET_ACCESS_SETTINGS
    ) -> EncordClientDataset:
        """
        Create and initialize a Encord client from a Encord config instance.

        Args:
            config: A Encord config instance.
            dataset_access_settings: Set the dataset_access_settings if you would like to change the defaults.

        Returns:
            EncordClientDataset: An Encord client dataset instance.
        """
        querier = Querier(config, resource_id=config.resource_id)
        key_type = querier.basic_getter(ApiKeyMeta)

        if key_type.resource_type == TYPE_PROJECT:
            raise RuntimeError("Trying to initialise an EncordClientDataset using a project key.")

        elif key_type.resource_type == TYPE_DATASET:
            logger.info("Initialising Encord client for dataset using key: %s", key_type.title)
            return EncordClientDataset(querier, config, dataset_access_settings=dataset_access_settings)

        else:
            raise encord.exceptions.InitialisationError(
                message=f"API key [{config.api_key}] is not associated with a project or dataset"
            )

    def get_dataset(self) -> OrmDataset:
        """
        Retrieve dataset info (pointers to data, labels).

        Returns:
            OrmDataset: A dataset record instance.

        Raises:
            AuthorisationError: If the dataset API key is invalid.
            ResourceNotFoundError: If no dataset exists by the specified dataset EntityId.
            UnknownError: If an error occurs while retrieving the dataset.
        """

        res = self._querier.basic_getter(
            OrmDataset,
            payload={
                "dataset_access_settings": dataclasses.asdict(self._dataset_access_settings),
            },
        )

        for row in res.data_rows:
            row["_querier"] = self._querier
        return res

    def list_data_rows(
        self,
        title_eq: Optional[str] = None,
        title_like: Optional[str] = None,
        created_before: Optional[Union[str, datetime]] = None,
        created_after: Optional[Union[str, datetime]] = None,
        data_types: Optional[List[DataType]] = None,
        data_hashes: Optional[List[str]] = None,
    ) -> List[DataRow]:
        """
        Retrieve dataset rows (pointers to data, labels).

        Args:
            title_eq: optional exact title row filter
            title_like: optional fuzzy title row filter; SQL syntax
            created_before: optional datetime row filter
            created_after: optional datetime row filter
            data_types: optional data types row filter
            data_hashes: optional list of individual data unit hashes to include

        Returns:
            List[DataRow]: A list of DataRows object that match the filter

        Raises:
            AuthorisationError: If the dataset API key is invalid.
            ResourceNotFoundError: If no dataset exists by the specified dataset EntityId.
            UnknownError: If an error occurs while retrieving the dataset.
        """

        created_before = optional_datetime_to_iso_str("created_before", created_before)
        created_after = optional_datetime_to_iso_str("created_after", created_after)

        data_rows = cast(
            List[DataRow],
            self._querier.get_multiple(
                DataRows,
                payload={
                    "title_eq": title_eq,
                    "title_like": title_like,
                    "created_before": created_before,
                    "created_after": created_after,
                    "data_types": [data_type.to_upper_case_string() for data_type in data_types]
                    if data_types is not None
                    else None,
                    "dataset_access_settings": dataclasses.asdict(self._dataset_access_settings),
                    "data_hashes": data_hashes,
                },
            ),
        )

        for row in data_rows:
            row["_querier"] = self._querier

        return data_rows

    def set_access_settings(self, dataset_access_settings: DatasetAccessSettings) -> None:
        self._dataset_access_settings = dataset_access_settings

    def add_users(self, user_emails: List[str], user_role: DatasetUserRole) -> List[DatasetUser]:
        """
        This function is documented in :meth:`encord.project.Dataset.add_users`.
        """

        payload = {"user_emails": user_emails, "user_role": user_role}
        users = self._querier.basic_setter(DatasetUsers, self._querier.resource_id, payload=payload)

        return [DatasetUser.from_dict(user) for user in users]

    def list_groups(self, dataset_hash: uuid.UUID) -> Page[DatasetGroup]:
        return self._get_api_client().get(
            f"datasets/{dataset_hash}/groups", params=None, result_type=Page[DatasetGroup]
        )

    def add_groups(self, dataset_hash: str, group_hash: List[uuid.UUID], user_role: DatasetUserRole) -> None:
        payload = AddDatasetGroupsPayload(group_hash_list=group_hash, user_role=user_role)
        self._get_api_client().post(f"datasets/{dataset_hash}/groups", params=None, payload=payload, result_type=None)

    def remove_groups(self, dataset_hash: uuid.UUID, group_hash: List[uuid.UUID]) -> None:
        params = RemoveGroupsParams(group_hash_list=group_hash)
        self._get_api_client().delete(f"datasets/{dataset_hash}/groups", params=params, result_type=None)

    def upload_video(
        self,
        file_path: Union[str, Path],
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        title: Optional[str] = None,
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> Video:
        """
        This function is documented in :meth:`encord.dataset.Dataset.upload_video`.
        """
        if os.path.exists(file_path):
            signed_urls = upload_to_signed_url_list(
                [file_path], self._config, self._querier, Video, cloud_upload_settings=cloud_upload_settings
            )
            res = upload_video_to_encord(signed_urls[0], title, folder_uuid, self._querier)
            if res:
                logger.info("Upload complete.")
                return Video(res)
            else:
                raise encord.exceptions.EncordException(message="An error has occurred during video upload.")
        else:
            raise encord.exceptions.EncordException(message=f"{file_path} does not point to a file.")

    def create_image_group(
        self,
        file_paths: Iterable[Union[str, Path]],
        max_workers: Optional[int] = None,
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        title: Optional[str] = None,
        *,
        create_video: bool = True,
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> List[ImageGroup]:
        """
        This function is documented in :meth:`encord.dataset.Dataset.create_image_group`.
        """
        for file_path in file_paths:
            if not os.path.exists(file_path):
                raise encord.exceptions.EncordException(message=f"{file_path} does not point to a file.")

        successful_uploads = upload_to_signed_url_list(
            file_paths, self._config, self._querier, Images, cloud_upload_settings=cloud_upload_settings
        )
        if not successful_uploads:
            raise encord.exceptions.EncordException("All image uploads failed. Image group was not created.")
        upload_images_to_encord(successful_uploads, self._querier)

        image_hash_list = [successful_upload.get("data_hash") for successful_upload in successful_uploads]
        payload = {
            "image_group_title": title,
            "create_video": create_video,
        }
        if folder_uuid is not None:
            payload["folder_uuid"] = str(folder_uuid)

        res = self._querier.basic_setter(
            ImageGroup,
            uid=image_hash_list,  # type: ignore
            payload=payload,
        )

        if res:
            titles = [video_data.get("title") for video_data in res]
            logger.info(f"Upload successful! {titles} created.")
            return [ImageGroup(obj) for obj in res]
        else:
            raise encord.exceptions.EncordException(message="An error has occurred during image group creation.")

    def create_dicom_series(
        self,
        file_paths: typing.Collection[Union[Path, str]],
        title: Optional[str] = None,
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> Dict:
        """
        This function is documented in :meth:`encord.dataset.Dataset.create_dicom_series`.
        """
        for file_path in file_paths:
            if not os.path.exists(file_path):
                raise encord.exceptions.EncordException(message=f"{file_path} does not point to a file.")

        successful_uploads = upload_to_signed_url_list(
            file_paths=file_paths,
            config=self._config,
            querier=self._querier,
            orm_class=DicomSeries,
            cloud_upload_settings=cloud_upload_settings,
        )
        if not successful_uploads:
            raise encord.exceptions.EncordException("DICOM files upload failed. The DICOM series was not created.")

        dicom_files = [
            {
                "id": file["data_hash"],
                "uri": file["file_link"],
                "title": file["title"],
            }
            for file in successful_uploads
        ]

        payload = {
            "title": title,
        }

        if folder_uuid is not None:
            payload["folder_uuid"] = str(folder_uuid)

        res = self._querier.basic_setter(DicomSeries, uid=dicom_files, payload=payload)
        if not res:
            raise encord.exceptions.EncordException(message="An error has occurred during the DICOM series creation.")

        return res

    def upload_image(
        self,
        file_path: Union[Path, str],
        title: Optional[str] = None,
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> Image:
        """
        This function is documented in :meth:`encord.dataset.Dataset.upload_image`.
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)
        if not file_path.is_file():
            raise encord.exceptions.EncordException(message=f"{str(file_path)} does not point to a file.")

        successful_uploads = upload_to_signed_url_list(
            [str(file_path)], self._config, self._querier, Images, cloud_upload_settings=cloud_upload_settings
        )
        if not successful_uploads:
            raise encord.exceptions.EncordException("Image upload failed.")

        upload = successful_uploads[0]
        if folder_uuid is not None:
            upload["folder_uuid"] = str(folder_uuid)
        if title is not None:
            upload["title"] = title

        res = self._querier.basic_setter(SingleImage, uid=None, payload=upload)

        if res["success"]:
            return Image({"data_hash": upload["data_hash"], "title": upload["title"], "file_link": upload["file_link"]})
        else:
            raise encord.exceptions.EncordException("Image upload failed.")

    def link_items(
        self,
        item_uuids: List[uuid.UUID],
        duplicates_behavior: DataLinkDuplicatesBehavior = DataLinkDuplicatesBehavior.SKIP,
    ) -> List[DataRow]:
        """
        Link storage items to the dataset, creating new data rows.

        Args:
            item_uuids: List of item UUIDs to link to the dataset
            duplicates_behaviour: The behavior to follow when encountering duplicates. Defaults to `SKIP`. See also
                :class:`encord.orm.dataset.DataLinkDuplicatesBehavior`
        """

        data_row_dicts = self._querier.basic_setter(
            DatasetLinkItems,
            uid=self._querier.resource_id,
            payload={
                "item_uuids": [str(item_uuid) for item_uuid in item_uuids],
                "duplicates_behavior": duplicates_behavior.value,
            },
        )
        return DataRow.from_dict_list(data_row_dicts)

    def delete_image_group(self, data_hash: str):
        """
        This function is documented in :meth:`encord.dataset.Dataset.delete_image_group`.
        """
        self._querier.basic_delete(ImageGroup, uid=data_hash)

    def delete_data(self, data_hashes: Union[List[str], str]):
        """
        This function is documented in :meth:`encord.dataset.Dataset.delete_data`.
        """
        if isinstance(data_hashes, str):
            data_hashes = [data_hashes]
        self._querier.basic_delete(Video, uid=data_hashes)

    def add_private_data_to_dataset(
        self,
        integration_id: str,
        private_files: Union[str, typing.Dict, Path, typing.TextIO],
        ignore_errors: bool = False,
    ) -> AddPrivateDataResponse:
        """
        This function is documented in :meth:`encord.dataset.Dataset.add_private_data_to_dataset`.
        """
        upload_job_id = self.add_private_data_to_dataset_start(
            integration_id,
            private_files,
            ignore_errors,
        )

        res = self.add_private_data_to_dataset_get_result(upload_job_id)

        if res.status == LongPollingStatus.DONE:
            return AddPrivateDataResponse(dataset_data_list=res.data_hashes_with_titles)
        elif res.status == LongPollingStatus.ERROR:
            raise encord.exceptions.EncordException(f"add_private_data_to_dataset errors occurred {res.errors}")
        else:
            raise ValueError(f"res.status={res.status}, this should never happen")

    def add_private_data_to_dataset_start(
        self,
        integration_id: str,
        private_files: Union[str, typing.Dict, Path, typing.TextIO],
        ignore_errors: bool = False,
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> str:
        """
        This function is documented in :meth:`encord.dataset.Dataset.add_private_data_to_dataset_start`.
        """
        if isinstance(private_files, dict):
            files = private_files
        elif isinstance(private_files, str):
            if os.path.exists(private_files):
                text_contents = Path(private_files).read_text(encoding="utf-8")
            else:
                text_contents = private_files

            files = json.loads(text_contents)
        elif isinstance(private_files, Path):
            text_contents = private_files.read_text(encoding="utf-8")
            files = json.loads(text_contents)
        elif isinstance(private_files, typing.TextIO):
            text_contents = private_files.read()
            files = json.loads(text_contents)
        else:
            raise ValueError(f"Type [{type(private_files)}] of argument private_files is not supported")

        payload = {
            "files": files,
            "integration_id": integration_id,
            "ignore_errors": ignore_errors,
        }
        if folder_uuid is not None:
            payload["folder_uuid"] = str(folder_uuid)

        process_hash = self._querier.basic_setter(
            DatasetDataLongPolling,
            self._querier.resource_id,
            payload=payload,
        )["process_hash"]

        logger.info(f"add_private_data_to_dataset job started with upload_job_id={process_hash}.")
        logger.info("SDK process can be terminated, this will not affect successful job execution.")
        logger.info("You can follow the progress in the web app via notifications.")

        return process_hash

    def add_private_data_to_dataset_get_result(
        self,
        upload_job_id: str,
        timeout_seconds: int = 7 * 24 * 60 * 60,  # 7 days
    ) -> DatasetDataLongPolling:
        """
        This function is documented in :meth:`encord.dataset.Dataset.add_private_data_to_dataset_get_result`.
        """
        failed_requests_count = 0
        polling_start_timestamp = time.perf_counter()

        while True:
            try:
                polling_elapsed_seconds = ceil(time.perf_counter() - polling_start_timestamp)
                polling_available_seconds = max(0, timeout_seconds - polling_elapsed_seconds)

                res = self._querier.basic_getter(
                    DatasetDataLongPolling,
                    self._querier.resource_id,
                    payload={
                        "process_hash": upload_job_id,
                        "timeout_seconds": min(
                            polling_available_seconds,
                            LONG_POLLING_MAX_REQUEST_TIME_SECONDS,
                        ),
                    },
                )

                if res.status == LongPollingStatus.DONE:
                    logger.info(f"add_private_data_to_dataset job completed with upload_job_id={upload_job_id}.")

                polling_elapsed_seconds = ceil(time.perf_counter() - polling_start_timestamp)
                polling_available_seconds = max(0, timeout_seconds - polling_elapsed_seconds)

                if polling_available_seconds == 0 or res.status in [LongPollingStatus.DONE, LongPollingStatus.ERROR]:
                    return res

                files_finished_count = res.units_done_count + res.units_error_count
                files_total_count = res.units_pending_count + res.units_done_count + res.units_error_count

                if files_finished_count != files_total_count:
                    logger.info(f"Processed {files_finished_count}/{files_total_count} files")
                else:
                    logger.info("Processed all files, dataset data linking and task creation is performed, please wait")

                failed_requests_count = 0
            except (requests.exceptions.RequestException, encord.exceptions.RequestException):
                failed_requests_count += 1

                if failed_requests_count >= LONG_POLLING_RESPONSE_RETRY_N:
                    raise

                time.sleep(LONG_POLLING_SLEEP_ON_FAILURE_SECONDS)

    def update_data_item(self, data_hash: str, new_title: str) -> bool:
        """This function is documented in :meth:`encord.dataset.Dataset.update_data_item`."""
        payload = [{"video_hash": data_hash, "title": new_title}]

        response = self._querier.basic_setter(Video, self.get_dataset().dataset_hash, payload=payload)
        try:
            return response.get("success", False)
        except AttributeError:
            return False

    def re_encode_data(self, data_hashes: List[str]):
        """
        This function is documented in :meth:`encord.dataset.Dataset.re_encode_data`.
        """
        payload = {"data_hash": data_hashes}
        return self._querier.basic_put(ReEncodeVideoTask, uid=None, payload=payload)

    def re_encode_data_status(self, job_id: int):
        """
        This function is documented in :meth:`encord.dataset.Dataset.re_encode_data_status`.
        """
        return self._querier.basic_getter(ReEncodeVideoTask, uid=job_id)

    def run_ocr(self, image_group_id: str) -> List[ImageGroupOCR]:
        """
        This function is documented in :meth:`encord.dataset.Dataset.run_ocr`.
        """

        payload = {"image_group_data_hash": image_group_id}
        return self._querier.get_multiple(ImageGroupOCR, payload=payload)


class EncordClientProject(EncordClient):
    """
    DEPRECATED - prefer using the :class:`encord.project.Project` instead
    """

    @property
    def project_hash(self) -> str:
        assert self._querier.resource_id, "Resource id can't be empty for created project client"
        return self._querier.resource_id  # type: ignore[attr-defined]

    def get_project(self, include_labels_metadata=True) -> OrmProject:
        """
        Retrieve project info (pointers to data, labels).

        Args:
            include_labels_metadata: if false, label row metadata information will not be returned.

        Returns:
            OrmProject: A project record instance.

        Raises:
            AuthorisationError: If the project API key is invalid.
            ResourceNotFoundError: If no project exists by the specified project EntityId.
            UnknownError: If an error occurs while retrieving the project.
        """
        return self._querier.basic_getter(
            OrmProject, payload={"include_labels_metadata": include_labels_metadata}, retryable=True
        )

    def get_project_v2(self) -> ProjectOrmV2:
        """
        This is an internal method, do not use it directly.
        Use :meth:`UserClient.get_project` instead.
        """
        return self._get_api_client().get(f"/projects/{self.project_hash}", params=None, result_type=ProjectOrmV2)

    def list_label_rows(
        self,
        edited_before: Optional[Union[str, datetime]] = None,
        edited_after: Optional[Union[str, datetime]] = None,
        label_statuses: Optional[List[AnnotationTaskStatus]] = None,
        shadow_data_state: Optional[ShadowDataState] = None,
        *,
        include_uninitialised_labels: bool = False,
        include_workflow_graph_node: bool = True,
        include_client_metadata: bool = False,
        include_images_data: bool = False,
        label_hashes: Optional[Union[List[str], List[UUID]]] = None,
        data_hashes: Optional[Union[List[str], List[UUID]]] = None,
        data_title_eq: Optional[str] = None,
        data_title_like: Optional[str] = None,
        workflow_graph_node_title_eq: Optional[str] = None,
        workflow_graph_node_title_like: Optional[str] = None,
        include_all_label_branches: bool = False,
    ) -> List[LabelRowMetadata]:
        """
        This function is documented in :meth:`encord.project.Project.list_label_rows`.
        """

        data_hashes = [str(data_hash) for data_hash in data_hashes] if data_hashes is not None else None
        label_hashes = [str(label_hash) for label_hash in label_hashes] if label_hashes is not None else None

        label_statuses_values = (
            [label_status.value for label_status in label_statuses] if label_statuses is not None else None
        )
        edited_before = optional_datetime_to_iso_str("edited_before", edited_before)
        edited_after = optional_datetime_to_iso_str("edited_after", edited_after)

        payload = {
            "edited_before": edited_before,
            "edited_after": edited_after,
            "label_statuses": label_statuses_values,
            "shadow_data_state": shadow_data_state.value if shadow_data_state else None,
            "include_uninitialised_labels": include_uninitialised_labels,
            "data_hashes": data_hashes,
            "label_hashes": label_hashes,
            "data_title_eq": data_title_eq,
            "data_title_like": data_title_like,
            "workflow_graph_node_title_eq": workflow_graph_node_title_eq,
            "workflow_graph_node_title_like": workflow_graph_node_title_like,
            "include_client_metadata": include_client_metadata,
            "include_images_data": include_images_data,
            "include_workflow_graph_node": include_workflow_graph_node,
            "include_all_label_branches": include_all_label_branches,
        }
        return self._querier.get_multiple(LabelRowMetadata, payload=payload, retryable=True)

    def set_label_status(self, label_hash: str, label_status: LabelStatus) -> bool:
        """
        This function is documented in :meth:`encord.project.Project.set_label_status`.
        """
        payload = {
            "label_status": label_status.value,
        }
        return self._querier.basic_setter(LabelStatus, label_hash, payload)

    def add_users(self, user_emails: List[str], user_role: ProjectUserRole) -> List[ProjectUser]:
        """
        This function is documented in :meth:`encord.project.Project.add_users`.
        """

        payload = {"user_emails": user_emails, "user_role": user_role}
        users = self._querier.basic_setter(ProjectUsers, self._querier.resource_id, payload=payload)

        return [ProjectUser.from_dict(user) for user in users]

    def list_groups(self, project_hash: uuid.UUID) -> Page[ProjectGroup]:
        return self._get_api_client().get(
            f"projects/{project_hash}/groups", params=None, result_type=Page[ProjectGroup]
        )

    def add_groups(self, project_hash: uuid.UUID, group_hash: List[uuid.UUID], user_role: ProjectUserRole) -> None:
        payload = AddProjectGroupsPayload(group_hash_list=group_hash, user_role=user_role)
        self._get_api_client().post(f"projects/{project_hash}/groups", params=None, payload=payload, result_type=None)

    def remove_groups(self, group_hash: List[uuid.UUID]) -> None:
        params = RemoveGroupsParams(group_hash_list=group_hash)
        self._get_api_client().delete(f"projects/{self.project_hash}/groups", params=params, result_type=None)

    def copy_project(
        self,
        copy_datasets: Union[bool, CopyDatasetOptions] = False,
        copy_collaborators=False,
        copy_models=False,
        *,
        copy_labels: Optional[CopyLabelsOptions] = None,
        new_title: Optional[str] = None,
        new_description: Optional[str] = None,
    ) -> str:
        """
        This function is documented in :meth:`encord.project.Project.copy_project`.
        """
        payload = CopyProjectPayload()
        payload.project_copy_metadata = CopyProjectPayload._ProjectCopyMetadata(new_title)
        payload.copy_labels_options = CopyProjectPayload._CopyLabelsOptions()

        if payload.project_copy_metadata and new_description:
            payload.project_copy_metadata.project_description = new_description

        if copy_labels:
            payload.copy_project_options.append(ProjectCopyOptions.LABELS)
            if isinstance(copy_labels, CopyLabelsOptions):
                payload.copy_labels_options.accepted_label_hashes = copy_labels.accepted_label_hashes
                payload.copy_labels_options.accepted_label_statuses = copy_labels.accepted_label_statuses

        if copy_datasets:
            if copy_datasets is True or copy_datasets.action == CopyDatasetAction.ATTACH:
                payload.copy_project_options.append(ProjectCopyOptions.DATASETS)
            elif copy_datasets.action == CopyDatasetAction.CLONE:
                payload.project_copy_metadata.dataset_title = copy_datasets.dataset_title
                payload.project_copy_metadata.dataset_description = copy_datasets.dataset_description
                payload.copy_labels_options.create_new_dataset = True
                payload.copy_labels_options.datasets_to_data_hashes_map = copy_datasets.datasets_to_data_hashes_map

        if copy_models:
            payload.copy_project_options.append(ProjectCopyOptions.MODELS)
        if copy_collaborators:
            payload.copy_project_options.append(ProjectCopyOptions.COLLABORATORS)

        return self._querier.basic_setter(ProjectCopy, self._querier.resource_id, payload=dataclasses.asdict(payload))

    def get_label_row(
        self,
        uid: str,
        get_signed_url: bool = True,
        *,
        include_object_feature_hashes: Optional[typing.Set[str]] = None,
        include_classification_feature_hashes: Optional[typing.Set[str]] = None,
        include_reviews: bool = False,
        include_export_history: bool = False,
    ) -> LabelRow:
        """
        This function is documented in :meth:`encord.project.Project.get_label_row`.
        """
        payload = {
            "get_signed_url": get_signed_url,
            "multi_request": False,
            "include_object_feature_hashes": optional_set_to_list(include_object_feature_hashes),
            "include_classification_feature_hashes": optional_set_to_list(include_classification_feature_hashes),
            "include_reviews": include_reviews,
            "include_export_history": include_export_history,
        }

        return self._querier.basic_getter(LabelRow, uid, payload=payload, retryable=True)

    def get_label_rows(
        self,
        uids: List[str],
        get_signed_url: bool = True,
        *,
        include_object_feature_hashes: Optional[typing.Set[str]] = None,
        include_classification_feature_hashes: Optional[typing.Set[str]] = None,
        include_reviews: bool = False,
        include_export_history: bool = False,
    ) -> List[LabelRow]:
        """
        This function is documented in :meth:`encord.project.Project.get_label_rows`.
        """
        payload = {
            "get_signed_url": get_signed_url,
            "multi_request": True,
            "include_object_feature_hashes": optional_set_to_list(include_object_feature_hashes),
            "include_classification_feature_hashes": optional_set_to_list(include_classification_feature_hashes),
            "include_reviews": include_reviews,
            "include_export_history": include_export_history,
        }

        return self._querier.get_multiple(LabelRow, uids, payload=payload, retryable=True)

    def save_label_row(self, uid, label, validate_before_saving: bool = False):
        """
        This function is documented in :meth:`encord.project.Project.save_label_row`.
        """
        label = LabelRow(label)
        if validate_before_saving:
            label["validate_before_saving"] = True
        return self._querier.basic_setter(LabelRow, uid, payload=label, retryable=True)

    def save_label_rows(self, uids: List[str], payload: List[LabelRow], validate_before_saving: bool = False):
        """
        This function is meant for internal use, please consider using :class:`encord.objects.LabelRowV2` class instead

        Saves multiple label rows. See :meth:`.save_label_row`

        Args:
            uids: list of label hashes
            payload: list of LabelRow objects

        Returns:
            None
        """
        multirequest_payload = {
            "multi_request": True,
            "labels": payload,
            "validate_before_saving": validate_before_saving,
        }
        return self._querier.basic_setter(LabelRow, uid=uids, payload=multirequest_payload, retryable=True)

    def create_label_row(self, uid, *, get_signed_url=False) -> typing.Any:
        """
        This function is documented in :meth:`encord.project.Project.create_label_row`.
        """
        return self._querier.basic_put(LabelRow, uid=uid, payload={"get_signed_url": get_signed_url})

    def create_label_rows(self, uids: List[str], *, get_signed_url=False) -> List[LabelRow]:
        """
        This function is meant for internal use, please consider using :class:`encord.objects.LabelRowV2` class instead

        Create multiple label rows. See :meth:`.create_label_row`

        Args:
            uids: list of data uids where label_status is NOT_LABELLED.

        Returns:
            List[LabelRow]: A list of created label rows
        """
        return self._querier.put_multiple(
            LabelRow, uid=uids, payload={"multi_request": True, "get_signed_url": get_signed_url}
        )

    def submit_label_row_for_review(self, uid):
        """
        This function is documented in :meth:`encord.project.Project.submit_label_row_for_review`.
        """
        return self._querier.basic_put(Review, uid=uid, payload=None)

    def add_datasets(self, dataset_hashes: List[str]) -> bool:
        """
        This function is documented in :meth:`encord.project.Project.add_datasets`.
        """
        payload = {"dataset_hashes": dataset_hashes}
        return self._querier.basic_setter(ProjectDataset, uid=None, payload=payload)

    def remove_datasets(self, dataset_hashes: List[str]) -> bool:
        """
        This function is documented in :meth:`encord.project.Project.remove_datasets`.
        """
        return self._querier.basic_delete(ProjectDataset, uid=dataset_hashes)

    def list_project_datasets(self, project_hash: UUID) -> Iterable[ProjectDataset]:
        return (
            self._get_api_client()
            .get(f"projects/{project_hash}/datasets", params=None, result_type=Page[ProjectDataset])
            .results
        )

    def get_project_ontology(self) -> LegacyOntology:
        project = self.get_project()
        ontology = project["editor_ontology"]
        return LegacyOntology.from_dict(ontology)

    @deprecated("0.1.102", alternative="encord.ontology.Ontology class")
    def add_object(self, name: str, shape: ObjectShape) -> bool:
        """
        This function is documented in :meth:`encord.project.Project.add_object`.
        """
        if not name:
            raise ValueError("Ontology object name is empty")

        ontology = self.get_project_ontology()
        ontology.add_object(name, shape)
        return self.__set_project_ontology(ontology)

    @deprecated("0.1.102", alternative="encord.ontology.Ontology class")
    def add_classification(
        self,
        name: str,
        classification_type: ClassificationType,
        required: bool,
        options: Optional[Iterable[str]] = None,
    ):
        """
        This function is documented in :meth:`encord.project.Project.add_classification`.
        """
        if not name:
            raise ValueError("Ontology classification name is empty")

        ontology = self.get_project_ontology()
        ontology.add_classification(name, classification_type, required, options)
        return self.__set_project_ontology(ontology)

    def list_models(self) -> List[ModelConfiguration]:
        """
        This function is documented in :meth:`encord.project.Project.list_models`.
        """
        return self._querier.get_multiple(ModelConfiguration)

    def get_training_metadata(
        self,
        model_iteration_uids: Iterable[str],
        get_created_at: bool = False,
        get_training_final_loss: bool = False,
        get_model_training_labels: bool = False,
    ) -> List[TrainingMetadata]:
        """
        This function is documented in :meth:`encord.project.Project.get_training_metadata`.
        """
        payload = {
            "model_iteration_uids": list(model_iteration_uids),
            "get_created_at": get_created_at,
            "get_training_final_loss": get_training_final_loss,
            "get_model_training_labels": get_model_training_labels,
        }
        return self._querier.get_multiple(TrainingMetadata, payload=payload)

    def create_model_row(
        self,
        title: str,
        description: str,
        features: List[str],
        model: Union[AutomationModels, str],
    ) -> str:
        """
        This function is documented in :meth:`encord.project.Project.create_model_row`.
        """
        if title is None:
            raise encord.exceptions.EncordException(message="You must set a title to create a model row.")

        if features is None:
            raise encord.exceptions.EncordException(
                message="You must pass a list of feature uid's (hashes) to create a model row."
            )

        if isinstance(model, AutomationModels):
            model = model.value

        elif model is None or not AutomationModels.has_value(model):  # Backward compatibility with string options
            raise encord.exceptions.EncordException(
                message="You must pass a model from the `encord.constants.model.AutomationModels` Enum to create a "
                "model row."
            )

        model_row = ModelRow(
            {
                "title": title,
                "description": description,
                "features": features,
                "model": model,
            }
        )

        model_payload = Model(
            {
                "model_operation": ModelOperations.CREATE.value,
                "model_parameters": model_row,
            }
        )

        return self._querier.basic_put(Model, None, payload=model_payload)

    def model_delete(self, uid: str) -> bool:
        """
        This function is documented in :meth:`encord.project.Project.model_delete`.
        """

        return self._querier.basic_delete(Model, uid=uid)

    def model_inference(
        self,
        uid: str,
        file_paths: Optional[List[str]] = None,
        base64_strings: Optional[List[bytes]] = None,
        conf_thresh: float = 0.6,
        iou_thresh: float = 0.3,
        device: Device = Device.CUDA,
        detection_frame_range: Optional[List[int]] = None,
        allocation_enabled: bool = False,
        data_hashes: Optional[List[str]] = None,
        rdp_thresh: float = 0.005,
    ):
        """
        This function is documented in :meth:`encord.project.Project.model_inference`.
        """
        if (file_paths is None and base64_strings is None and data_hashes is None) or (
            file_paths is not None
            and len(file_paths) > 0
            and base64_strings is not None
            and len(base64_strings) > 0
            and data_hashes is not None
            and len(data_hashes) > 0
        ):
            raise encord.exceptions.EncordException(
                message="To run model inference, you must pass either a list of files or base64 strings or list of"
                " data hash."
            )

        if detection_frame_range is None:
            detection_frame_range = []

        files = []
        if file_paths is not None:
            for file_path in file_paths:
                file = open(file_path, "rb").read()
                files.append(
                    {
                        "uid": file_path,  # Add file path as inference identifier
                        "base64_str": base64.b64encode(file).decode("utf-8"),  # File to base64 string
                    }
                )

        elif base64_strings is not None:
            for base64_string in base64_strings:
                files.append(
                    {
                        "uid": str(uuid.uuid4()),  # Add uuid as inference identifier
                        "base64_str": base64_string.decode("utf-8"),  # base64 string to utf-8
                    }
                )

        inference_params = ModelInferenceParams(
            {
                "files": files,
                "conf_thresh": conf_thresh,
                "iou_thresh": iou_thresh,
                "device": _device_to_string(device),
                "detection_frame_range": detection_frame_range,
                "allocation_enabled": allocation_enabled,
                "data_hashes": data_hashes,
                "rdp_thresh": rdp_thresh,
            }
        )

        model = Model(
            {
                "model_operation": ModelOperations.INFERENCE.value,
                "model_parameters": inference_params,
            }
        )

        return self._querier.basic_setter(Model, uid, payload=model)

    def model_train_start(
        self,
        model_hash: Union[str, UUID],
        label_rows: List[Union[str, UUID]],
        epochs: int,
        weights: ModelTrainingWeights,
        batch_size: int = 24,
        device: Device = Device.CUDA,
    ) -> UUID:
        """
        This function is documented in :meth:`encord.project.Project.model_train_start`.
        """

        if not label_rows:
            raise encord.exceptions.EncordException(
                message="You must pass a list of label row uid's (hashes) to train a model."
            )

        if not epochs:
            raise encord.exceptions.EncordException(message="You must set number of epochs to train a model.")

        if not batch_size:
            raise encord.exceptions.EncordException(message="You must set a batch size to train a model.")

        if weights is None or not isinstance(weights, ModelTrainingWeights):
            raise encord.exceptions.EncordException(
                message="You must pass weights from the `encord.constants.model_weights` module to train a model."
            )

        training_hash = self._get_api_client().post(
            f"ml-models/{model_hash}/training",
            params=None,
            payload=PublicModelTrainStartPayload(
                label_rows=[x if isinstance(x, UUID) else UUID(x) for x in label_rows],
                epochs=epochs,
                batch_size=batch_size,
                model=weights.model,
                training_weights_link=weights.training_weights_link,
                device=_device_to_string(device),
            ),
            result_type=UUID,
        )

        logger.info(f"model_train job started with training_hash={training_hash}.")
        logger.info("SDK process can be terminated, this will not affect successful job execution.")
        logger.info("You can follow the progress in the SDK using model_train_get_result method.")
        logger.info("You can also follow the progress in the web app via notifications.")

        return training_hash

    def model_train_get_result(
        self,
        model_hash: Union[str, UUID],
        training_hash: Union[str, UUID],
        timeout_seconds: int = 7 * 24 * 60 * 60,  # 7 days
    ) -> dict:
        """
        This function is documented in :meth:`encord.project.Project.model_train_get_result`.
        """

        failed_requests_count = 0
        polling_start_timestamp = time.perf_counter()

        while True:
            try:
                polling_elapsed_seconds = ceil(time.perf_counter() - polling_start_timestamp)
                polling_available_seconds = max(0, timeout_seconds - polling_elapsed_seconds)

                logger.info(f"__model_train_get_result started polling call {polling_elapsed_seconds=}")
                tmp_res = self._get_api_client().get(
                    f"ml-models/{model_hash}/{training_hash}/training",
                    params=PublicModelTrainGetResultParams(
                        timeout_seconds=min(
                            polling_available_seconds,
                            LONG_POLLING_MAX_REQUEST_TIME_SECONDS,
                        ),
                    ),
                    result_type=PublicModelTrainGetResultResponse,
                )

                if tmp_res.status == PublicModelTrainGetResultLongPollingStatus.DONE:
                    logger.info(f"model_train job completed with training_hash={training_hash}.")

                polling_elapsed_seconds = ceil(time.perf_counter() - polling_start_timestamp)
                polling_available_seconds = max(0, timeout_seconds - polling_elapsed_seconds)

                if polling_available_seconds == 0 or tmp_res.status in [
                    PublicModelTrainGetResultLongPollingStatus.DONE,
                    PublicModelTrainGetResultLongPollingStatus.ERROR,
                ]:
                    res = tmp_res
                    break

                failed_requests_count = 0
            except (requests.exceptions.RequestException, encord.exceptions.RequestException):
                failed_requests_count += 1

                if failed_requests_count >= LONG_POLLING_RESPONSE_RETRY_N:
                    raise

                time.sleep(LONG_POLLING_SLEEP_ON_FAILURE_SECONDS)

        if res.status == PublicModelTrainGetResultLongPollingStatus.DONE:
            if res.result is None:
                raise ValueError(f"{res.status=}, res.result should not be None with DONE status")

            return res.result.dict()
        elif res.status == PublicModelTrainGetResultLongPollingStatus.ERROR:
            raise encord.exceptions.EncordException("model_train error occurred")
        else:
            raise ValueError(
                f"res.status={res.status}, only DONE and ERROR status is expected after successful long polling"
            )

    def object_interpolation(
        self,
        key_frames,
        objects_to_interpolate,
    ):
        """
        This function is documented in :meth:`encord.project.Project.object_interpolation`.
        """
        if len(key_frames) == 0 or len(objects_to_interpolate) == 0:
            raise encord.exceptions.EncordException(
                message="To run object interpolation, you must pass key frames and objects to interpolate.."
            )

        interpolation_params = ObjectInterpolationParams(
            {
                "key_frames": key_frames,
                "objects_to_interpolate": objects_to_interpolate,
            }
        )

        algo = LabelingAlgorithm(
            {
                "algorithm_name": INTERPOLATION,
                "algorithm_parameters": interpolation_params,
            }
        )

        return self._querier.basic_setter(LabelingAlgorithm, str(uuid.uuid4()), payload=algo)

    def fitted_bounding_boxes(
        self,
        frames: dict,
        video: dict,
    ):
        """
        This function is documented in :meth:`encord.project.Project.fitted_bounding_boxes`.
        """
        if len(frames) == 0 or len(video) == 0:
            raise encord.exceptions.EncordException(
                message="To run fitting, you must pass frames and video to run bounding box fitting on.."
            )

        fitting_params = BoundingBoxFittingParams(
            {
                "labels": frames,
                "video": video,
            }
        )

        algo = LabelingAlgorithm(
            {
                "algorithm_name": FITTED_BOUNDING_BOX,
                "algorithm_parameters": fitting_params,
            }
        )

        return self._querier.basic_setter(LabelingAlgorithm, str(uuid.uuid4()), payload=algo)

    def get_data(self, data_hash: str, get_signed_url: bool = False) -> Tuple[Optional[Video], Optional[List[Image]]]:
        """
        This function is documented in :meth:`encord.project.Project.get_data`.
        """
        uid = {"data_hash": data_hash, "get_signed_url": get_signed_url}

        dataset_data: DatasetData = self._querier.basic_getter(DatasetData, uid=uid)

        video: Union[Video, None] = None
        if dataset_data["video"] is not None:
            video = Video(dataset_data["video"])

        images: Union[List[Image], None] = None
        if dataset_data["images"] is not None:
            images = []

            for image in dataset_data["images"]:
                images.append(Image(image))

        return video, images

    def get_label_logs(
        self,
        user_hash: Optional[str] = None,
        data_hash: Optional[str] = None,
        from_unix_seconds: Optional[int] = None,
        to_unix_seconds: Optional[int] = None,
        after: Optional[datetime] = None,
        before: Optional[datetime] = None,
        user_email: Optional[str] = None,
    ) -> List[LabelLog]:
        """
        This function is documented in :meth:`encord.project.Project.get_label_logs`.
        """
        if after is not None:
            if from_unix_seconds is not None:
                raise ValueError("Only one of 'from_unix_seconds' and 'after' parameters should be specified")

            from_unix_seconds = int(after.timestamp())

        if before is not None:
            if to_unix_seconds is not None:
                raise ValueError("Only one of 'to_unix_seconds' and 'before' parameters should be specified")

            to_unix_seconds = int(before.timestamp())

        payload = LabelLogParams(
            user_hash=user_hash,
            user_email=user_email,
            data_hash=data_hash,
            end_timestamp=to_unix_seconds,
            start_timestamp=from_unix_seconds,
        )

        return self._querier.get_multiple(LabelLog, payload=payload.to_dict(by_alias=False))

    def __set_project_ontology(self, ontology: LegacyOntology) -> bool:
        """
        Save updated project ontology
        Args:
            ontology: the updated project ontology

        Returns:
            bool
        """
        payload = {"editor": ontology.to_dict()}
        return self._querier.basic_setter(OrmProject, uid=None, payload=payload)

    def workflow_reopen(self, label_hashes: List[str]) -> None:
        """
        This function is documented in :meth:`encord.objects.LabelRowV2.workflow_reopen`.
        """
        self._querier.basic_setter(
            LabelWorkflowGraphNode,
            label_hashes,
            payload=LabelWorkflowGraphNodePayload(action=WorkflowAction.REOPEN),
        )

    def workflow_complete(self, label_hashes: List[str]) -> None:
        """
        This function is documented in :meth:`encord.objects.LabelRowV2.workflow_complete`.
        """
        self._querier.basic_setter(
            LabelWorkflowGraphNode,
            label_hashes,
            payload=LabelWorkflowGraphNodePayload(action=WorkflowAction.COMPLETE),
        )

    def workflow_set_priority(self, priorities: List[Tuple[str, float]]) -> None:
        self._get_api_client().post(
            f"projects/{self.project_hash}/priorities",
            params=None,
            payload=TaskPriorityParams(priorities=priorities),
            result_type=None,
        )

    def get_collaborator_timers(self, params: CollaboratorTimerParams) -> Iterable[CollaboratorTimer]:
        yield from self._get_api_client().get_paged_iterator(
            "analytics/collaborators/timers", params=params, result_type=CollaboratorTimer
        )

    def get_label_validation_errors(self, label_hash: str) -> List[str]:
        errors = self._get_api_client().get(
            f"projects/{self.project_hash}/labels/{label_hash}/validation-state",
            params=None,
            result_type=LabelValidationState,
        )

        if errors.is_valid:
            return []

        return errors.errors or []


def _device_to_string(device: Device) -> str:
    if not isinstance(device, Device):
        if device is None or not Device.has_value(device):  # Backward compatibility with string options
            raise EncordException(message="You must pass a device from the `from encord.constants.model.Device` enum.")
        return cast(str, device)

    return device.value


CordClientProject = EncordClientProject



import json
from datetime import datetime
from functools import wraps
from typing import Any, Callable, Dict, Generic, Literal, Type, TypeVar

from pydantic import BaseModel, Field, PrivateAttr, ValidationError, root_validator, validator
from pydantic.generics import GenericModel
from pydantic.json import pydantic_encoder
from typing_extensions import Self

from encord.common.time_parser import parse_datetime
from encord.common.utils import snake_to_camel
from encord.exceptions import EncordException
from encord.orm.base_dto.base_dto_interface import BaseDTOInterface, T


class BaseDTO(BaseDTOInterface, BaseModel):
    class Config:
        ignore_extra = True
        alias_generator = snake_to_camel
        allow_population_by_field_name = True

    @validator("*", pre=True)
    def parse_datetime(cls, value, field):
        if isinstance(value, str) and issubclass(field.type_, datetime):
            return parse_datetime(value)
        return value

    @classmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        try:
            return cls.parse_obj(d)  # type: ignore[attr-defined]
        except ValidationError as e:
            raise EncordException(message=str(e)) from e

    def to_dict(self, by_alias=True, exclude_none=True) -> Dict[str, Any]:
        # Pydantic v1 is missing the 'model_dump()' method, the below is suboptimal but works
        return json.loads(self.json(by_alias=by_alias, exclude_none=exclude_none))  # type: ignore[attr-defined]


DataT = TypeVar("DataT")


class GenericBaseDTO(BaseDTOInterface, GenericModel):
    class Config:
        ignore_extra = True
        alias_generator = snake_to_camel
        allow_population_by_field_name = True

    @validator("*", pre=True)
    def parse_datetime(cls, value, field):
        if isinstance(value, str) and issubclass(field.type_, datetime):
            return parse_datetime(value)
        return value

    @classmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        try:
            return cls.parse_obj(d)  # type: ignore[attr-defined]
        except ValidationError as e:
            raise EncordException(message=str(e)) from e

    def to_dict(self, by_alias=True, exclude_none=True) -> Dict[str, Any]:
        # Pydantic v1 is missing the 'model_dump()' method, the below is suboptimal but works
        return json.loads(self.json(by_alias=by_alias, exclude_none=exclude_none))  # type: ignore[attr-defined]


def dto_validator(mode: Literal["before", "after"] = "before") -> Callable:
    def decorator(func: Callable) -> Callable:
        return root_validator(pre=(mode == "before"))(func)  # type: ignore

    return decorator


RootType = TypeVar("RootType")


class RootModelDTO(GenericModel, Generic[RootType]):
    __root__: RootType

    def __init__(cls, *args, **kwargs):
        if "__root__" in kwargs:
            super().__init__(*args, **kwargs)
        else:
            root = args[0] if len(args) > 0 else kwargs.pop("root", None)
            super().__init__(__root__=root)

    @property
    def root(self) -> Type:
        return self.__root__  # type: ignore[return-value]

    @classmethod
    def model_validate(cls, *args, **kwargs) -> Self:
        return cls.validate(*args, **kwargs)

    def model_dump(self, *args, **kwargs) -> dict:
        mode = kwargs.pop("mode")
        value = self.dict(*args, **kwargs)
        raw_value = value.get("__root__")
        if mode == "json":
            return json.loads(json.dumps(raw_value, default=pydantic_encoder))
        else:
            return raw_value  # type: ignore[return-value]



from collections import OrderedDict
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
from typing import List, Optional, Union
from uuid import UUID

from encord.common.time_parser import parse_datetime
from encord.constants.model import AutomationModels
from encord.exceptions import EncordException
from encord.orm import base_orm
from encord.orm.analytics import CamelStrEnum
from encord.orm.base_dto import BaseDTO
from encord.orm.formatter import Formatter
from encord.utilities.common import ENCORD_CONTACT_SUPPORT_EMAIL


class ModelOperations(Enum):
    INFERENCE = 0
    TRAIN = 1
    CREATE = 2


class Model(base_orm.BaseORM):
    """
    Model base ORM.

    ORM:

    model_operation,
    model_parameters,

    """

    DB_FIELDS = OrderedDict([("model_operation", int), ("model_parameters", dict)])


@dataclass
class ModelConfiguration(Formatter):
    model_uid: str
    title: str
    description: str
    feature_node_hashes: List[str]
    """The corresponding feature node hashes of the ontology object"""
    model: AutomationModels
    model_iteration_uids: List[str]
    """All the UIDs of individual model training instances"""

    @classmethod
    def from_dict(cls, json_dict: dict):
        return ModelConfiguration(
            model_uid=json_dict["model_uid"],
            title=json_dict["title"],
            description=json_dict["description"],
            feature_node_hashes=cls._get_feature_node_hashes(json_dict["feature_node_hashes"]),
            model=cls._get_automation_model(json_dict["model"]),
            model_iteration_uids=json_dict["model_iteration_uids"],
        )

    @staticmethod
    def _get_feature_node_hashes(features: dict) -> List[str]:
        return list(features.keys())

    @staticmethod
    def _get_automation_model(automation_model_str: str) -> AutomationModels:
        try:
            return AutomationModels(automation_model_str)
        except ValueError as e:
            raise EncordException(
                "A model was returned which was not recognised. Please upgrade your SDK "
                f"to the latest version or contact support at {ENCORD_CONTACT_SUPPORT_EMAIL}."
            ) from e


@dataclass
class ModelTrainingLabelMetadata:
    label_uid: str
    data_uid: str
    data_link: str

    @staticmethod
    def from_dict(json_dict: dict) -> "ModelTrainingLabelMetadata":
        return ModelTrainingLabelMetadata(
            label_uid=json_dict["label_uid"],
            data_uid=json_dict["data_uid"],
            data_link=json_dict["data_link"],
        )

    @classmethod
    def from_list(cls, json_list: list) -> List["ModelTrainingLabelMetadata"]:
        return [cls.from_dict(item) for item in json_list]


@dataclass
class ModelTrainingLabel:
    label_metadata_list: List[ModelTrainingLabelMetadata]
    feature_uids: List[str]

    @staticmethod
    def from_dict(json_dict: dict) -> "ModelTrainingLabel":
        return ModelTrainingLabel(
            label_metadata_list=ModelTrainingLabelMetadata.from_list(json_dict["label_metadata_list"]),
            feature_uids=json_dict["feature_uids"],
        )


@dataclass
class TrainingMetadata(Formatter):
    model_iteration_uid: str
    created_at: Optional[datetime] = None
    training_final_loss: Optional[float] = None
    model_training_labels: Optional[ModelTrainingLabel] = None

    @classmethod
    def from_dict(cls, json_dict: dict):
        return TrainingMetadata(
            model_iteration_uid=json_dict["model_iteration_uid"],
            created_at=cls.get_created_at(json_dict),
            training_final_loss=json_dict["training_final_loss"],
            model_training_labels=cls.get_model_training_labels(json_dict),
        )

    @staticmethod
    def get_created_at(json_dict: dict) -> Optional[datetime]:
        created_at = json_dict["created_at"]
        if created_at is None:
            return None

        return parse_datetime(created_at)

    @staticmethod
    def get_model_training_labels(json_dict: dict) -> Optional[ModelTrainingLabel]:
        model_training_labels = json_dict["model_training_labels"]
        if model_training_labels is None:
            return None

        return ModelTrainingLabel.from_dict(model_training_labels)


class ModelRow(base_orm.BaseORM):
    """
    A model row contains a set of features and a model (resnet18, resnet34, resnet50, resnet101, resnet152,
    vgg16, vgg19, faster_rcnn, mask_rcnn).

    ORM:

    model_hash (uid),
    title,
    description,
    features,
    model,

    """

    DB_FIELDS = OrderedDict(
        [
            ("model_hash", str),
            ("title", str),
            ("description", str),
            ("features", list),
            ("model", str),
        ]
    )


class ModelInferenceParams(base_orm.BaseORM):
    """
    Model inference parameters for running models trained via the platform.

    ORM:

    local_file_path,
    conf_thresh,
    iou_thresh,
    device
    detection_frame_range (optional)

    """

    DB_FIELDS = OrderedDict(
        [
            ("files", list),
            ("conf_thresh", float),  # Confidence threshold
            ("iou_thresh", float),  # Intersection over union threshold
            ("device", str),
            ("detection_frame_range", list),
            ("allocation_enabled", bool),
            ("data_hashes", list),
            ("rdp_thresh", float),
        ]
    )


class ModelTrainingWeights(base_orm.BaseORM):
    """
    Model training weights.

    ORM:

    training_weights_link,

    """

    DB_FIELDS = OrderedDict(
        [
            ("model", str),
            ("training_weights_link", str),
        ]
    )


class ModelTrainingParams(base_orm.BaseORM):
    """
    Model training parameters.

    ORM:

    model_hash,
    label_rows,
    epochs,
    batch_size,
    weights,
    device

    """

    DB_FIELDS = OrderedDict(
        [
            ("model_hash", str),
            ("label_rows", list),
            ("epochs", int),
            ("batch_size", int),
            ("weights", ModelTrainingWeights),
            ("device", str),
        ]
    )


class PublicModelTrainStartPayload(BaseDTO):
    label_rows: List[UUID]
    epochs: int
    batch_size: int
    model: str
    training_weights_link: Optional[str] = None
    device: str


class PublicModelTrainGetResultParams(BaseDTO):
    timeout_seconds: int


class PublicModelTrainGetResultLongPollingStatus(str, Enum):
    DONE = "DONE"
    ERROR = "ERROR"
    PENDING = "PENDING"


class PublicModelTrainGetResultResponseDoneResult(BaseDTO):
    model_hash: UUID
    training_hash: UUID
    title: str
    type: str
    model: str
    framework: str
    epochs: int
    batch_size: int
    final_loss: float
    weights_link: str
    created_at: datetime
    duration: int


class PublicModelTrainGetResultResponse(BaseDTO):
    status: PublicModelTrainGetResultLongPollingStatus
    result: Optional[PublicModelTrainGetResultResponseDoneResult] = None



from datetime import datetime
from enum import Enum, auto
from typing import Optional

from encord.common.utils import snake_to_camel
from encord.orm.base_dto import BaseDTO
from encord.utilities.project_user import ProjectUserRole


class CamelStrEnum(str, Enum):
    # noinspection PyMethodParameters
    def _generate_next_value_(name, start, count, last_values) -> str:  # type: ignore
        return snake_to_camel(name)


class CollaboratorTimersGroupBy(CamelStrEnum):
    DATA_UNIT = auto()
    PROJECT = auto()


class CollaboratorTimerParams(BaseDTO):
    project_hash: str
    after: datetime
    before: Optional[datetime] = None
    group_by: CollaboratorTimersGroupBy = CollaboratorTimersGroupBy.DATA_UNIT
    page_size: int = 100
    page_token: Optional[str] = None


class CollaboratorTimer(BaseDTO):
    user_email: str
    user_role: ProjectUserRole
    data_title: Optional[str] = None
    time_seconds: float



"""
---
title: "Model Weights"
slug: "sdk-ref-model-weights"
hidden: false
metadata:
  title: "Model Weights"
  description: "Encord SDK Model Weights."
category: "64e481b57b6027003f20aaa0"
---
"""

from encord.constants.model import AutomationModels
from encord.orm.model import ModelTrainingWeights

fast_ai = ModelTrainingWeights(
    {
        "model": AutomationModels.FAST_AI.value,
        "training_weights_link": None,
    }
)

faster_rcnn_R_50_C4_1x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_721ade.pkl",
    }
)

faster_rcnn_R_50_DC5_1x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_51d356.pkl",
    }
)

faster_rcnn_R_50_FPN_1x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_b275ba.pkl",
    }
)

faster_rcnn_R_50_C4_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_f97cb7.pkl",
    }
)

faster_rcnn_R_50_DC5_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_68d202.pkl",
    }
)

faster_rcnn_R_50_FPN_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_280758.pkl",
    }
)

faster_rcnn_R_101_C4_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_298dad.pkl",
    }
)

faster_rcnn_R_101_DC5_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_3e0943.pkl",
    }
)

faster_rcnn_R_101_FPN_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_f6e8b1.pkl",
    }
)

faster_rcnn_X_101_32x8d_FPN_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.FASTER_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_68b088.pkl",
    }
)

mask_rcnn_X_101_32x8d_FPN_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.MASK_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_2d9806.pkl",
    }
)

mask_rcnn_R_50_C4_1x = ModelTrainingWeights(
    {
        "model": AutomationModels.MASK_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_9243eb.pkl",
    }
)

mask_rcnn_R_50_C4_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.MASK_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_4ce675.pkl",
    }
)

mask_rcnn_R_101_FPN_3x = ModelTrainingWeights(
    {
        "model": AutomationModels.MASK_RCNN.value,
        "training_weights_link": "https://storage.googleapis.com/aws_model_backup/model_final_a3ec72.pkl",
    }
)



"""
---
title: "Objects - Classification"
slug: "sdk-ref-objects-classification"
hidden: false
metadata:
  title: "Objects - Classification"
  description: "Encord SDK Objects - Classification."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Type, TypeVar

from encord.objects.attributes import (
    Attribute,
    AttributeType,
    _add_attribute,
    attribute_from_dict,
    attributes_to_list_dict,
)
from encord.objects.ontology_element import OntologyElement


@dataclass
class Classification(OntologyElement):
    """
    Represents a whole-image classification as part of the Ontology structure. Wraps a single Attribute that describes
    the image in general rather than an individual object.

    Attributes:
        uid (int): The unique identifier for the classification.
        feature_node_hash (str): A unique hash identifying the feature node.
        attributes (List[Attribute]): A list of attributes associated with this classification.
    """

    uid: int
    feature_node_hash: str
    attributes: List[Attribute]

    @property
    def title(self) -> str:
        """
        Returns the title of the classification, which is the name of the first attribute.

        Returns:
            str: The title of the classification.
        """
        return self.attributes[0].name

    @property
    def children(self) -> Sequence[OntologyElement]:
        """
        Returns the attributes of the classification as children elements.

        Returns:
            Sequence[OntologyElement]: The attributes of the classification.
        """
        return self.attributes

    def create_instance(self) -> ClassificationInstance:
        """
        Create a ClassificationInstance to be used with a label row.

        Returns:
            ClassificationInstance: An instance of ClassificationInstance.
        """
        return ClassificationInstance(self)

    @classmethod
    def from_dict(cls, d: dict) -> Classification:
        """
        Create a Classification instance from a dictionary.

        Args:
            d (dict): A dictionary containing classification information.

        Returns:
            Classification: An instance of Classification.
        """
        attributes_ret: List[Attribute] = [attribute_from_dict(attribute_dict) for attribute_dict in d["attributes"]]
        return Classification(
            uid=int(d["id"]),
            feature_node_hash=d["featureNodeHash"],
            attributes=attributes_ret,
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the Classification instance to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the classification.

        Raises:
            ValueError: If the classification does not have any attributes.
        """
        ret: Dict[str, Any] = {
            "id": str(self.uid),
            "featureNodeHash": self.feature_node_hash,
        }
        if attributes_list := attributes_to_list_dict(self.attributes):
            ret["attributes"] = attributes_list
        else:
            raise ValueError(f"Classification {str(self.uid)} requires attribute before use")

        return ret

    def add_attribute(
        self,
        cls: Type[AttributeType],
        name: str,
        local_uid: Optional[int] = None,
        feature_node_hash: Optional[str] = None,
        required: bool = False,
    ) -> AttributeType:
        """
        Adds an attribute to the classification.

        Args:
            cls (Type[AttributeType]): The attribute type, one of `RadioAttribute`, `ChecklistAttribute`, `TextAttribute`.
            name (str): The user-visible name of the attribute.
            local_uid (Optional[int]): Integer identifier of the attribute. Normally auto-generated; omit this unless
                                       the aim is to create an exact clone of an existing ontology.
            feature_node_hash (Optional[str]): Global identifier of the attribute. Normally auto-generated; omit this
                                               unless the aim is to create an exact clone of an existing ontology.
            required (bool): Whether the label editor would mark this attribute as 'required'.

        Returns:
            AttributeType: The created attribute that can be further specified with Options, where appropriate.

        Raises:
            ValueError: If the classification already has an attribute assigned.
        """
        if self.attributes:
            raise ValueError("Classification should have exactly one root attribute")
        return _add_attribute(self.attributes, cls, name, [self.uid], local_uid, feature_node_hash, required)

    def __hash__(self):
        return hash(self.feature_node_hash)


from encord.objects.classification_instance import ClassificationInstance



import dataclasses
import logging
from contextlib import contextmanager
from typing import Any, Dict, Generator, List, Optional, Sequence, Tuple, Type, TypeVar, Union

import orjson
import requests
import requests.exceptions
from requests import Session
from requests.adapters import HTTPAdapter, Retry

from encord.configs import ApiKeyConfig, BaseConfig
from encord.exceptions import RequestException, ResourceNotFoundError
from encord.http.common import (
    HEADER_CLOUD_TRACE_CONTEXT,
    RequestContext,
)
from encord.http.error_utils import check_error_response
from encord.http.query_methods import QueryMethods
from encord.http.request import Request, UIDType
from encord.orm.base_dto import BaseDTO
from encord.orm.formatter import Formatter

logger = logging.getLogger(__name__)


class Querier:
    """Querier for DB get/post requests."""

    T = TypeVar("T")
    PayloadType = Union[None, Dict[str, Any], BaseDTO, Sequence[Dict[str, Any]], Sequence[BaseDTO]]

    def __init__(self, config: BaseConfig, resource_type: Optional[str] = None, resource_id: Optional[str] = None):
        self._config = config
        self.resource_type = resource_type
        self.resource_id = resource_id
        if resource_id is None and isinstance(config, ApiKeyConfig):
            self.resource_id = config.resource_id

    def basic_getter(
        self, db_object_type: Type[T], uid: UIDType = None, payload: PayloadType = None, retryable=True
    ) -> T:
        """Single DB object getter."""
        request = self._request(QueryMethods.GET, db_object_type, uid, self._config.read_timeout, payload=payload)
        res, context = self._execute(request, retryable=retryable)
        if res:
            return self._parse_response(db_object_type, res)
        else:
            raise ResourceNotFoundError("Resource not found.", context=context)

    def get_multiple(
        self, object_type: Type[T], uid: UIDType = None, payload: PayloadType = None, retryable=True
    ) -> List[T]:
        return self._request_multiple(QueryMethods.GET, object_type, uid, payload, retryable=retryable)

    def post_multiple(
        self, object_type: Type[T], uid: UIDType = None, payload: PayloadType = None, retryable=False
    ) -> List[T]:
        return self._request_multiple(QueryMethods.POST, object_type, uid, payload)

    def put_multiple(
        self, object_type: Type[T], uid: UIDType = None, payload: PayloadType = None, retryable=False
    ) -> List[T]:
        return self._request_multiple(QueryMethods.PUT, object_type, uid, payload)

    def _request_multiple(
        self, method: QueryMethods, object_type: Type[T], uid: UIDType, payload: PayloadType = None, retryable=False
    ) -> List[T]:
        request = self._request(method, object_type, uid, self._config.read_timeout, payload=payload)
        result, context = self._execute(request, retryable=retryable)

        if result is not None:
            return [self._parse_response(object_type, item) for item in result]
        else:
            raise ResourceNotFoundError(
                f"[{object_type}] not found for query with uid=[{uid}] and payload=[{payload}]", context=context
            )

    @staticmethod
    def _parse_response(object_type: Type[T], item: Dict[str, Any]) -> T:
        if issubclass(object_type, BaseDTO):
            return object_type.from_dict(item)  # type: ignore
        if issubclass(object_type, Formatter):
            return object_type.from_dict(item)  # type: ignore
        elif dataclasses.is_dataclass(object_type):
            return object_type(**item)  # type: ignore
        else:
            return object_type(item)  # type: ignore

    @staticmethod
    def _serialise_payload(payload: PayloadType) -> Union[None, Dict[str, Any], List[Dict[str, Any]]]:
        if isinstance(payload, BaseDTO):
            return payload.to_dict()
        elif isinstance(payload, Sequence) and all(isinstance(x, BaseDTO) for x in payload):
            return [x.to_dict() for x in payload]  # type: ignore
        else:
            return payload  # type: ignore

    def basic_delete(self, db_object_type: Type[T], uid: UIDType = None, retryable=False):
        """Single DB object getter."""
        request = self._request(
            QueryMethods.DELETE,
            db_object_type,
            uid,
            self._config.read_timeout,
        )

        res, _ = self._execute(request, retryable=retryable)
        return res

    def basic_setter(self, db_object_type: Type[T], uid: UIDType, payload: PayloadType, retryable=False):
        """Single DB object setter."""
        request = self._request(
            QueryMethods.POST,
            db_object_type,
            uid,
            self._config.write_timeout,
            payload=payload,
        )

        res, context = self._execute(request, retryable=retryable)

        if res is not None:
            return res
        else:
            raise RequestException(f"Setting {db_object_type} with uid {uid} failed.", context=context)

    def basic_put(self, db_object_type, uid, payload: PayloadType, retryable: bool = True, enable_logging: bool = True):
        """Single DB object put request."""
        request = self._request(
            QueryMethods.PUT,
            db_object_type,
            uid,
            self._config.write_timeout,
            payload=payload,
        )

        res, context = self._execute(request, retryable=retryable, enable_logging=enable_logging)

        if res:
            return res
        else:
            raise RequestException(f"Setting {db_object_type} with uid {uid} failed.", context=context)

    @staticmethod
    def _exception_context(request: requests.PreparedRequest) -> RequestContext:
        try:
            x_cloud_trace_context = request.headers.get(HEADER_CLOUD_TRACE_CONTEXT)
            if x_cloud_trace_context is None:
                return RequestContext()

            x_cloud_trace_context = x_cloud_trace_context.split(";")[0]
            trace_id, span_id = (x_cloud_trace_context.split("/") + [None, None])[:2]
            return RequestContext(trace_id=trace_id, span_id=span_id)
        except Exception:
            return RequestContext()

    def _request(
        self, method: QueryMethods, db_object_type: Type[T], uid: UIDType, timeout: int, payload: PayloadType = None
    ):
        request = Request(
            method, db_object_type, uid, timeout, self._config.connect_timeout, self._serialise_payload(payload)
        )

        request.headers = self._config.define_headers(
            resource_id=self.resource_id, resource_type=self.resource_type, data=request.data
        )
        return request

    def _execute(self, request: Request, retryable=False, enable_logging: bool = True) -> Tuple[Any, RequestContext]:
        """Execute a request."""
        if enable_logging:
            logger.info("Request: %s", (request.data[:100] + "..") if len(request.data) > 100 else request.data)

        req = requests.Request(
            method=str(request.http_method),
            url=self._config.endpoint,
            headers=request.headers,
            data=request.data,
        ).prepare()

        context = self._exception_context(req)

        timeouts = (request.connect_timeout, request.timeout)

        req_settings = self._config.requests_settings
        with create_new_session(
            max_retries=req_settings.max_retries if retryable else 0,
            backoff_factor=req_settings.backoff_factor,
            connect_retries=req_settings.connection_retries,
        ) as session:
            try:
                res = session.send(req, timeout=timeouts)
            except Exception as e:
                raise RequestException(f"Request session.send failed {req.method=} {req.url=}", context=context) from e

            try:
                res_json = orjson.loads(res.content)
            except Exception as e:
                raise RequestException(f"Error parsing JSON response: {res.text.strip()}", context=context) from e

        # pylint: disable-next=no-member
        if res_json.get("status") != requests.codes.ok:
            response = res_json.get("response")
            extra_payload = res_json.get("payload")
            check_error_response(response, context, extra_payload)

        return res_json.get("response"), context


@contextmanager
def create_new_session(
    max_retries: Optional[int], backoff_factor: float, connect_retries
) -> Generator[Session, None, None]:
    retry_policy = Retry(
        connect=connect_retries,
        read=max_retries,
        status=max_retries,  # type: ignore
        other=max_retries,  # type: ignore
        allowed_methods=["POST", "PUT", "GET"],  # type: ignore  # post is there since we use it for idempotent ops too.
        status_forcelist=[413, 429, 500, 502, 503],
        backoff_factor=backoff_factor,
    )

    with Session() as session:
        session.mount("http://", HTTPAdapter(max_retries=retry_policy))
        session.mount("https://", HTTPAdapter(max_retries=retry_policy))

        yield session



"""
---
title: "Objects - Skeleton Objects"
slug: "sdk-ref-objects-skelly"
hidden: false
metadata:
  title: "Objects - Skeleton Objects"
  description: "Encord SDK Objects - Skeleton Objects."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Optional, Set, Type

from encord.objects.coordinates import SkeletonCoordinate, SkeletonCoordinates
from encord.orm.skeleton_template import SkeletonTemplate as SkeletonTemplateORM
from encord.orm.skeleton_template import SkeletonTemplateCoordinate


class SkeletonTemplate(SkeletonTemplateORM):
    @property
    def required_vertices(self) -> Set[str]:
        """
        Get the set of required vertex names for the skeleton.

        Returns:
            Set[str]: A set containing the names of the required vertices.
        """
        return {coordinate.name for coordinate in self.skeleton.values()}

    def create_instance(self, provided_coordinates: List[SkeletonCoordinate]) -> SkeletonCoordinates:
        """
        Create an instance of SkeletonCoordinates with the provided coordinates.

        Args:
            provided_coordinates (List[SkeletonCoordinate]): A list of SkeletonCoordinate objects to align.

        Returns:
            SkeletonCoordinates: An instance of SkeletonCoordinates with aligned coordinates.

        Raises:
            ValueError: If the provided vertices do not match the required vertices.
        """
        provided_vertices = {provided_coordinate.name for provided_coordinate in provided_coordinates}
        if provided_vertices != self.required_vertices:
            difference = provided_vertices.symmetric_difference(self.required_vertices)
            raise ValueError(f"Provided vertices does not match the required vertices, difference is {difference}")
        aligned_coordinates = []
        for coord in provided_coordinates:
            partner = [x for x in self.skeleton.values() if x.name == coord.name][0]
            aligned_coordinate = SkeletonCoordinate(
                x=coord.x, y=coord.y, name=coord.name, feature_hash=partner.feature_hash
            )
            aligned_coordinates.append(aligned_coordinate)
        return SkeletonCoordinates(values=aligned_coordinates, name=self.name)



"""
---
title: "Ontology"
slug: "sdk-ref-ontology"
hidden: false
metadata:
  title: "Ontology"
  description: "Encord SDK Ontology class"
category: "64e481b57b6027003f20aaa0"
---
"""

import datetime
from typing import Iterable, List, Union
from uuid import UUID

from encord.http.querier import Querier
from encord.http.v2.api_client import ApiClient
from encord.http.v2.payloads import Page
from encord.objects.ontology_structure import OntologyStructure
from encord.orm.group import AddOntologyGroupsPayload, OntologyGroup, RemoveGroupsParams
from encord.orm.ontology import Ontology as OrmOntology
from encord.utilities.hash_utilities import convert_to_uuid
from encord.utilities.ontology_user import OntologyUserRole


class Ontology:
    """
    Access ontology related data and manipulate the ontology. Instantiate this class via
    :meth:`encord.user_client.EncordUserClient.get_ontology()`
    """

    def __init__(self, querier: Querier, instance: OrmOntology, api_client: ApiClient):
        self._querier = querier
        self._ontology_instance = instance
        self.api_client = api_client

    @property
    def ontology_hash(self) -> str:
        """
        Get the ontology hash (i.e. the Ontology ID).
        """
        return self._ontology_instance.ontology_hash

    @property
    def title(self) -> str:
        """
        Get the title of the ontology.
        """
        return self._ontology_instance.title

    @title.setter
    def title(self, value: str) -> None:
        self._ontology_instance.title = value

    @property
    def description(self) -> str:
        """
        Get the description of the ontology.
        """
        return self._ontology_instance.description

    @description.setter
    def description(self, value: str) -> None:
        self._ontology_instance.description = value

    @property
    def created_at(self) -> datetime.datetime:
        """
        Get the time the ontology was created at.
        """
        return self._ontology_instance.created_at

    @property
    def last_edited_at(self) -> datetime.datetime:
        """
        Get the time the ontology was last edited at.
        """
        return self._ontology_instance.last_edited_at

    @property
    def structure(self) -> OntologyStructure:
        """
        Get the structure of the ontology.
        """
        return self._ontology_instance.structure

    def refetch_data(self) -> None:
        """
        The Ontology class will only fetch its properties once. Use this function if you suspect the state of those
        properties to be dirty.
        """
        self._ontology_instance = self._get_ontology()

    def save(self) -> None:
        """
        Sync local state to the server, if updates are made to structure, title or description fields
        """
        if self._ontology_instance:
            payload = dict(**self._ontology_instance)
            payload["editor"] = self._ontology_instance.structure.to_dict()  # we're using internal/legacy name here
            payload.pop("structure", None)
            self._querier.basic_put(OrmOntology, self._ontology_instance.ontology_hash, payload)

    def _get_ontology(self):
        return self._querier.basic_getter(OrmOntology, self._ontology_instance.ontology_hash)

    def list_groups(self) -> Iterable[OntologyGroup]:
        """
        List all groups that have access to a particular ontology.
        """
        ontology_hash = convert_to_uuid(self.ontology_hash)
        page = self.api_client.get(f"ontologies/{ontology_hash}/groups", params=None, result_type=Page[OntologyGroup])

        yield from page.results

    def add_group(self, group_hash: Union[List[UUID], UUID], user_role: OntologyUserRole):
        """
        Add group to an ontology.

        Args:
            group_hash: List of group hashes to be added.
            user_role: User role that the group will be given.

        Returns:
            None
        """
        ontology_hash = convert_to_uuid(self.ontology_hash)
        if isinstance(group_hash, UUID):
            group_hash = [group_hash]
        payload = AddOntologyGroupsPayload(group_hash_list=group_hash, user_role=user_role)
        self.api_client.post(
            f"ontologies/{ontology_hash}/groups",
            params=None,
            payload=payload,
            result_type=None,
        )

    def remove_group(self, group_hash: Union[List[UUID], UUID]):
        """
        Remove group from ontology.

        Args:
            group_hash: List of group hashes to be removed.

        Returns:
            None
        """
        ontology_hash = convert_to_uuid(self.ontology_hash)
        if isinstance(group_hash, UUID):
            group_hash = [group_hash]
        params = RemoveGroupsParams(group_hash_list=group_hash)
        self.api_client.delete(f"ontologies/{ontology_hash}/groups", params=params, result_type=None)



"""
---
title: "Model"
slug: "sdk-ref-model"
hidden: false
metadata:
  title: "Model"
  description: "Encord SDK Model classes."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from enum import Enum
from typing import Set, cast


class Device(Enum):
    CUDA = "cuda"
    CPU = "cpu"

    @classmethod
    def has_value(cls, value) -> bool:
        # pylint: disable-next=no-member
        return value in cls._value2member_map_


class AutomationModels(Enum):
    FAST_AI = "fast_ai"
    RESNET18 = "resnet18"
    RESNET34 = "resnet34"
    RESNET50 = "resnet50"
    RESNET101 = "resnet101"
    RESNET152 = "resnet152"
    VGG16 = "vgg16"
    VGG19 = "vgg19"
    FASTER_RCNN = "faster_rcnn"
    MASK_RCNN = "mask_rcnn"

    @staticmethod
    def classification_options() -> Set[AutomationModels]:
        """
        Returns:
            Model types that can be used for frame-level classifications.
            That is, model types that can be used with ``<feature_node_hash>``es from
            the ``classification`` part of the ontology.
        """
        return {
            AutomationModels.FAST_AI,
            AutomationModels.RESNET18,
            AutomationModels.RESNET34,
            AutomationModels.RESNET50,
            AutomationModels.RESNET101,
            AutomationModels.RESNET152,
            AutomationModels.VGG16,
            AutomationModels.VGG19,
        }

    @staticmethod
    def object_detection_options() -> Set[AutomationModels]:
        """
        Returns:
            Model types that can be used with bounding_box type
            ``<feature_node_hashes>`` from the ``objects`` part of the project ontology.
        """
        return {AutomationModels.FASTER_RCNN}

    @staticmethod
    def instance_segmentation_options() -> Set[AutomationModels]:
        """
        Returns:
            Model types that can be used with polygon type ``<feature_node_hashes>``
            from the ``objects`` part of the project ontology.
        """
        return {AutomationModels.MASK_RCNN}

    @classmethod
    def has_value(cls, value) -> bool:
        # pylint: disable-next=no-member
        return value in cls._value2member_map_


# For backward compatibility
FAST_AI = cast(str, AutomationModels.FAST_AI.value)
RESNET18 = cast(str, AutomationModels.RESNET18.value)
RESNET34 = cast(str, AutomationModels.RESNET34.value)
RESNET50 = cast(str, AutomationModels.RESNET50.value)
RESNET101 = cast(str, AutomationModels.RESNET101.value)
RESNET152 = cast(str, AutomationModels.RESNET152.value)
VGG16 = cast(str, AutomationModels.VGG16.value)
VGG19 = cast(str, AutomationModels.VGG19.value)
FASTER_RCNN = cast(str, AutomationModels.FASTER_RCNN.value)
MASK_RCNN = cast(str, AutomationModels.MASK_RCNN.value)



"""
---
title: "Utilities - Project"
slug: "sdk-ref-utilities-project"
hidden: false
metadata:
  title: "Utilities - Project"
  description: "Encord SDK Utilities - Project."
category: "64e481b57b6027003f20aaa0"
---
"""

from typing import List, Set

from encord.orm.model import ModelConfiguration


def get_all_model_iteration_uids(model_configurations: List[ModelConfiguration]) -> Set[str]:
    """A convenience function that works well with :meth:`encord.project.Project.list_models`"""
    model_iteration_uids = set()
    for model_configuration in model_configurations:
        model_iteration_uids.update(model_configuration.model_iteration_uids)
    return model_iteration_uids



"""
---
title: "OntologyClassification DEPRECATED"
slug: "sdk-ref-ontology-classification-deprecated"
hidden: false
metadata:
  title: "OntologyClassification DEPRECATED"
  description: "Encord SDK OntologyClassification DEPRECATED."
category: "64e481b57b6027003f20aaa0"
---
"""

from dataclasses import dataclass
from typing import List

from encord.project_ontology.classification_attribute import ClassificationAttribute


@dataclass
class OntologyClassification:
    """
    DEPRECATED: prefer using :class:`encord.ontology.Ontology`

    A dataclass which holds classifications of the :class:`encord.project_ontology.Ontology`.
    """

    #: A unique (to the ontology) identifier of the classification.
    id: str
    #: An 8-character hex string uniquely defining the option.
    feature_node_hash: str
    #: A List of attributes for the classification.
    attributes: List[ClassificationAttribute]



from dataclasses import dataclass


@dataclass(frozen=True)
class ProjectWithUserRole:
    """
    This is a helper class denoting the relationship between the current user an a project
    """

    user_role: int
    project: dict



from __future__ import annotations

from datetime import datetime
from enum import IntEnum
from typing import Optional

from encord.common.deprecated import deprecated
from encord.orm.base_dto import BaseDTO


class Action(IntEnum):
    ADD = 0
    EDIT = 1
    DELETE = 2
    START = 3
    END = 4
    MARK_AS_NOT_LABELLED = 5
    MARK_AS_IN_PROGRESS = 6
    MARK_AS_LABELLED = 7
    MARK_AS_REVIEW_REQUIRED = 8
    MARK_AS_REVIEWED = 9
    MARK_AS_REVIEWED_TWICE = 10
    SUBMIT_TASK = 11
    APPROVE_LABEL = 12
    REJECT_LABEL = 13
    CLICK_SAVE = 14
    CLICK_UNDO = 15
    CLICK_REDO = 16
    CLICK_BULK = 17
    CLICK_ZOOM = 19
    CLICK_BRIGHTNESS = 20
    CLICK_HOTKEYS = 21
    CLICK_SETTINGS = 22
    ADD_ATTRIBUTE = 23
    EDIT_ATTRIBUTE = 24
    DELETE_ATTRIBUTE = 25
    APPROVE_NESTED_ATTRIBUTE = 26
    REJECT_NESTED_ATTRIBUTE = 27
    SUBMIT_LABEL = 28
    SUBMIT_NESTED_ATTRIBUTE = 29
    BUFFERING_OVERLAY_SHOWN = 30
    BITRATE_WARNING_SHOWN = 31
    SEEKING_OVERLAY_SHOWN = 32
    APPROVE_TASK = 33
    REJECT_TASK = 34
    CONSENSUS_APPROVE_LABEL = 35
    CONSENSUS_CANCEL_APPROVE_LABEL = 36
    CONSENSUS_IN_AGREEMENT = 37


class LabelLog(BaseDTO):
    log_hash: str
    user_hash: str
    user_email: str
    data_hash: str
    action: Action
    created_at: datetime
    identifier: Optional[str]
    feature_hash: Optional[str]
    label_name: Optional[str]
    time_taken: Optional[int]
    frame: Optional[int]

    @property
    @deprecated(version="0.1.100", alternative="LabelLog.identifier")
    def annotation_hash(self) -> Optional[str]:
        """
        DEPRECATED: this field is only provided for backwards compatibility, and will be removed in the future versions.
        Please use :attr:`identifier <LabelLog.identifier>` instead.
        """
        return self.identifier


class LabelLogParams(BaseDTO):
    user_hash: Optional[str]
    data_hash: Optional[str]
    start_timestamp: Optional[int]
    end_timestamp: Optional[int]
    user_email: Optional[str]
    # Flag for backwards compatibility
    include_user_email_and_interface_key: bool = True



"""
---
title: "Consensus Review Stage"
slug: "sdk-ref-stage-consen-review"
hidden: false
metadata:
  title: "Consensus Review Stage"
  description: "Encord SDK Consensus Review Stage."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from enum import Enum
from typing import Iterable, List, Literal, Optional, Union
from uuid import UUID

from encord.common.utils import ensure_list, ensure_uuid_list
from encord.http.bundle import Bundle
from encord.orm.base_dto import BaseDTO
from encord.orm.workflow import WorkflowStageType
from encord.workflow.common import TasksQueryParams, WorkflowAction, WorkflowStageBase, WorkflowTask


class ConsensusReviewTaskStatus(str, Enum):
    NEW = "NEW"
    ASSIGNED = "ASSIGNED"
    RELEASED = "RELEASED"
    REOPENED = "REOPENED"


class _ReviewTasksQueryParams(TasksQueryParams):
    user_emails: Optional[List[str]] = None
    data_hashes: Optional[List[UUID]] = None
    dataset_hashes: Optional[List[UUID]] = None
    data_title_contains: Optional[str] = None
    statuses: Optional[List[ConsensusReviewTaskStatus]] = None


class ConsensusReviewStage(WorkflowStageBase):
    stage_type: Literal[WorkflowStageType.CONSENSUS_REVIEW] = WorkflowStageType.CONSENSUS_REVIEW

    """
    The Review stage for Consensus Workflows.
    """

    def get_tasks(
        self,
        *,
        assignee: Union[List[str], str, None] = None,
        data_hash: Union[List[UUID], UUID, List[str], str, None] = None,
        dataset_hash: Union[List[UUID], UUID, List[str], str, None] = None,
        data_title: Optional[str] = None,
        status: Union[ConsensusReviewTaskStatus, List[ConsensusReviewTaskStatus], None] = None,
    ) -> Iterable[ConsensusReviewTask]:
        """
        Retrieves tasks for the ConsensusReviewStage.

        **Parameters**

        - `assignee` (Union[List[str], str, None]): A list of user emails or a single user email to filter tasks by assignee.
        - `data_hash` (Union[List[UUID], UUID, List[str], str, None]): A list of data unit UUIDs or a single data unit UUID to filter tasks by data hash.
        - `dataset_hash` (Union[List[UUID], UUID, List[str], str, None]): A list of dataset UUIDs or a single dataset UUID to filter tasks by dataset hash.
        - `data_title` (Optional[str]): A string to filter tasks by data title.
        - `status` (Union[ConsensusReviewTaskStatus, List[ConsensusReviewTaskStatus], None]): A list of task statuses or a single task status to filter tasks by their status.

        **Returns**

        An iterable of `ConsensusReviewTask` instances with the following information:
        - `uuid`: Unique identifier for the task.
        - `created_at`: Time and date the task was created.
        - `updated_at`: Time and date the task was last edited.
        - `assignee`: The user currently assigned to the task. The value is None if no one is assigned to the task.
        - `data_hash`: Unique identifier for the data unit.
        - `data_title`: Name/title of the data unit.
        - `options`: List of ConsensusReviewOptions. ConsensusReviewOptions are the labels available for each subtask, including information such as annotator, label_branch_name, and label_hash.
        """
        params = _ReviewTasksQueryParams(
            user_emails=ensure_list(assignee),
            data_hashes=ensure_uuid_list(data_hash),
            dataset_hashes=ensure_uuid_list(dataset_hash),
            data_title_contains=data_title,
            statuses=ensure_list(status),
        )

        for task in self._workflow_client.get_tasks(self.uuid, params, type_=ConsensusReviewTask):
            task._stage_uuid = self.uuid
            task._workflow_client = self._workflow_client
            yield task


class _ActionApprove(WorkflowAction):
    action: Literal["APPROVE"] = "APPROVE"


class _ActionReject(WorkflowAction):
    action: Literal["REJECT"] = "REJECT"


class _ActionAssign(WorkflowAction):
    action: Literal["ASSIGN"] = "ASSIGN"
    assignee: str


class _ActionRelease(WorkflowAction):
    action: Literal["RELEASE"] = "RELEASE"


class ConsensusReviewOption(BaseDTO):
    annotator: str
    label_branch_name: str
    label_hash: Optional[UUID]


class ConsensusReviewTask(WorkflowTask):
    status: ConsensusReviewTaskStatus
    data_hash: UUID
    data_title: str
    assignee: Optional[str]
    options: List[ConsensusReviewOption]

    """
    Represents tasks in the Review stage of a Consensus Project.

    **Attributes**

    - `status` (ConsensusReviewTaskStatus): The status of the review task.
    - `data_hash` (UUID): Unique ID for the data unit.
    - `data_title` (str): Name of the data unit.
    - `assignee` (Optional[str]): The user assigned to the task.
    - `options` (List[ConsensusReviewOption]): Specify the labels for the task.

    **Allowed actions**

    - `approve`: Approves a task.
    - `reject`: Rejects a task.
    - `assign`: Assigns a task to a user.
    - `release`: Releases a task from the current user.
    """

    def approve(self, *, bundle: Optional[Bundle] = None) -> None:
        """
        Approve the current task.

        **Parameters**

        - `bundle` (Optional[Bundle]): Optional bundle of actions to execute with the approval.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(stage_uuid, _ActionApprove(task_uuid=self.uuid), bundle=bundle)

    def reject(self, *, bundle: Optional[Bundle] = None) -> None:
        """
        Reject the current task.

        **Parameters**

        - `bundle` (Optional[Bundle]): Optional bundle of actions to execute with the rejection.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(stage_uuid, _ActionReject(task_uuid=self.uuid), bundle=bundle)

    def assign(self, assignee: str, *, bundle: Optional[Bundle] = None) -> None:
        """
        Assign the current task to a user.

        **Parameters**

        - `assignee` (str): The user to assign the task to.
        - `bundle` (Optional[Bundle]): Optional bundle of actions to execute with the assignment.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(stage_uuid, _ActionAssign(task_uuid=self.uuid, assignee=assignee), bundle=bundle)

    def release(self, *, bundle: Optional[Bundle] = None) -> None:
        """
        Release the current task from the current user.

        **Parameters**

        - `bundle` (Optional[Bundle]): Optional bundle of actions to execute with the release.

        **Returns**

        None
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(stage_uuid, _ActionRelease(task_uuid=self.uuid), bundle=bundle)



import importlib.metadata as importlib_metadata

from encord.orm.base_dto.base_dto_interface import BaseDTOInterface

pydantic_version_str = importlib_metadata.version("pydantic")

pydantic_version = int(pydantic_version_str.split(".")[0])
if pydantic_version < 2:
    from encord.orm.base_dto.base_dto_pydantic_v1 import (
        BaseDTO,
        Field,
        GenericBaseDTO,
        PrivateAttr,
        RootModelDTO,
        dto_validator,
    )
else:
    from encord.orm.base_dto.base_dto_pydantic_v2 import (  # type: ignore[assignment]
        BaseDTO,
        Field,
        GenericBaseDTO,
        PrivateAttr,
        RootModelDTO,
        dto_validator,
    )
